import datetime
from unittest import TestCase
import mock

from google.api_core.exceptions import BadRequest

import bigflow

from bigflow.bigquery.interactive import InteractiveDatasetManager
from bigflow.bigquery.interactive import interactive_component
from bigflow.bigquery.job import DEFAULT_RETRY_COUNT
from bigflow.bigquery.job import DEFAULT_RETRY_PAUSE_SEC
from bigflow.bigquery.interactive import log_syntax_error


class OperationLevelDatasetManagerTestCase(TestCase):

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_preserve_dataset_order(self, create_dataset_manager_mock):
        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, {'dependency_kwargs': kwargs})

        # given
        dataset1, dataset2, dataset3 = [
            InteractiveDatasetManager(
                project_id='project{}'.format(str(i)),
                dataset_name='dataset{}'.format(str(i)))
            for i in range(1, 4)
        ]

        @interactive_component(ds1=dataset1, ds2=dataset2, ds3=dataset3)
        def standard_component(ds2, ds1, ds3):
            # then
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['project_id'], 'project1')
            self.assertEqual(ds2._dataset_manager['dependency_kwargs']['project_id'], 'project2')
            self.assertEqual(ds3._dataset_manager['dependency_kwargs']['project_id'], 'project3')

        # when
        job = standard_component.to_job()

        # then
        job.execute(bigflow.JobContext.make())

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_pass_all_arguments_to_core_dataset_manager_on_run(self, create_dataset_manager_mock):
        # given
        fake_dataset_manager = mock.Mock()
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, fake_dataset_manager)

        dataset = InteractiveDatasetManager(
            project_id='project1',
            dataset_name='dataset1')

        @interactive_component(ds=dataset)
        def standard_component(ds):
            ds.write_truncate('some table', 'some sql', partitioned=False, custom_run_datetime='2019-12-12')
            ds.write_append('some table', 'some sql', partitioned=False, custom_run_datetime='2019-12-12')
            ds.write_tmp('some table', 'some sql', custom_run_datetime='2019-12-12')
            ds.collect('some sql', custom_run_datetime='2019-12-12')
            ds.dry_run('some sql', custom_run_datetime='2019-12-12')
            ds.create_table('some sql')
            ds.load_table_from_dataframe(
                'some table', 'some dataframe', partitioned=False, custom_run_datetime='2019-12-12')

        # when
        job = standard_component.to_job()
        job.execute(bigflow.JobContext.make())

        # then
        fake_dataset_manager.assert_has_calls([
            mock.call.write_truncate(
                table_name='some table', sql='some sql', partitioned=False, custom_run_datetime='2019-12-12'),
            mock.call.write_append(
                table_name='some table', sql='some sql', partitioned=False, custom_run_datetime='2019-12-12'),
            mock.call.write_tmp(
                table_name='some table', sql='some sql', custom_run_datetime='2019-12-12'),
            mock.call.collect(
                sql='some sql', custom_run_datetime='2019-12-12'),
            mock.call.dry_run(
                sql='some sql', custom_run_datetime='2019-12-12'),
            mock.call.create_table(create_query='some sql'),
            mock.call.load_table_from_dataframe(
                table_name='some table', df='some dataframe', partitioned=False, custom_run_datetime='2019-12-12')
        ])

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_run_specified_operation(self, create_dataset_manager_mock):
        # given
        dataset_manager_mock = mock.Mock()
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, dataset_manager_mock)

        default_dataset = InteractiveDatasetManager(
            project_id='project1',
            dataset_name='dataset1')

        @interactive_component(ds=default_dataset)
        def standard_component(ds):
            # then
            ds.write_truncate('table1', 'some sql', operation_name='operation1')
            ds.write_truncate('table2', 'some sql', operation_name='operation2')
            ds.write_truncate('table3', 'some sql', operation_name='operation3')

        # when
        standard_component.run('2019-01-01', operation_name='operation2')

        # then
        dataset_manager_mock.assert_has_calls([mock.call.write_truncate(
            table_name='table2',
            sql='some sql',
            custom_run_datetime=None,
            partitioned=True)])

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_peek_operation_results(self, create_dataset_manager_mock):
        # given
        dataset_manager_mock = mock.Mock()
        dataset_manager_mock.collect.side_effect = lambda sql, **kwargs: sql
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, dataset_manager_mock)

        default_dataset = InteractiveDatasetManager(
            project_id='project1',
            dataset_name='dataset1')

        @interactive_component(ds=default_dataset)
        def standard_component(ds):
            # then
            ds.write_truncate('table1', 'some sql', operation_name='operation1')

            ds.collect('collect sql 2', operation_name='operation2')
            ds.write_truncate('table3', 'collect sql 3', operation_name='operation3')
            ds.write_append('table4', 'collect sql 4', operation_name='operation4')
            ds.write_tmp('table5', 'collect sql 5', operation_name='operation5')

            ds.write_truncate('table3', 'some sql', operation_name='operation6')

        # expected
        self.assertEqual(
            standard_component.peek('2019-01-01', operation_name='operation2'),
            'collect sql 2\nLIMIT 1000')
        self.assertEqual(
            standard_component.peek('2019-01-01', operation_name='operation3'),
            'collect sql 3\nLIMIT 1000')
        self.assertEqual(
            standard_component.peek('2019-01-01', operation_name='operation4'),
            'collect sql 4\nLIMIT 1000')
        self.assertEqual(
            standard_component.peek('2019-01-01', operation_name='operation5'),
            'collect sql 5\nLIMIT 1000')


    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_allow_setting_peek_limit(self, create_dataset_manager_mock):
        # given
        dataset_manager_mock = mock.Mock()
        dataset_manager_mock.collect.side_effect = lambda sql, **kwargs: sql
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, dataset_manager_mock)

        default_dataset = InteractiveDatasetManager(
            project_id='project1',
            dataset_name='dataset1')

        @interactive_component(ds=default_dataset)
        def standard_component(ds):
            # then
            ds.write_truncate('table1', 'some sql', operation_name='operation1')
            ds.collect('collect sql', operation_name='operation2')
            ds.write_truncate('table3', 'some sql', operation_name='operation3')

        # when
        result = standard_component.peek('2019-01-01', operation_name='operation2', limit=1)

        # then
        self.assertEqual(result, 'collect sql\nLIMIT 1')


class InteractiveComponentToJobTestCase(TestCase):

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_pass_configuration_to_job(self, create_dataset_manager_mock):
        # given

        class DatasetManagerStub(object):
            def __init__(self, runtime_str, extras):
                self.runtime_str = runtime_str
                self.extras = extras
                self.client = 'client'

        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, DatasetManagerStub(
            kwargs.get('runtime'), kwargs.get('extras')))

        default_dataset = InteractiveDatasetManager(
            project_id='fake_project',
            dataset_name='fake_dataset',
            extras={'extra': 'param'})

        @interactive_component(ds=default_dataset)
        def standard_component(ds):
            # then
            self.assertEqual(ds.dt, '2019-01-01 00:00:00')
            self.assertEqual(ds.extras, {'extra': 'param'})
            self.assertEqual(ds.client, 'client')

        # when
        job = standard_component.to_job()

        # then
        self.assertEqual(job.id, 'standard_component')
        self.assertEqual(job.retry_count, DEFAULT_RETRY_COUNT)
        self.assertEqual(job.retry_pause_sec, DEFAULT_RETRY_PAUSE_SEC)
        self.assertEqual(job.dependency_configuration, {'ds': default_dataset.config})
        job.execute(bigflow.JobContext.make(runtime=datetime.datetime(2019, 1, 1)))

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_turn_component_to_job_with_redefined_dependencies(self, create_dataset_manager_mock):
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, {'dependency_kwargs': kwargs})

        # given
        default_dataset = InteractiveDatasetManager(
            project_id='default_project',
            dataset_name='default_dataset',
            internal_tables=['table1'],
            external_tables={'table': 'table1'},
            extras={'param': 'param1'})

        @interactive_component(ds1=default_dataset, ds2=default_dataset)
        def standard_component(ds1, ds2):
            # then
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['project_id'], 'default_project')
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['dataset_name'], 'default_dataset')
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['internal_tables'], ['table1'])
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['external_tables'], {'table': 'table1'})
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['extras'], {'param': 'param1'})

            self.assertEqual(ds2._dataset_manager['dependency_kwargs']['project_id'], 'modified_project')
            self.assertEqual(ds2._dataset_manager['dependency_kwargs']['dataset_name'], 'modified_dataset')
            self.assertEqual(ds2._dataset_manager['dependency_kwargs']['internal_tables'], ['table2'])
            self.assertEqual(ds2._dataset_manager['dependency_kwargs']['external_tables'], {'table': 'table2'})
            self.assertEqual(ds2._dataset_manager['dependency_kwargs']['extras'], {'param': 'param2'})

        modified_dataset = InteractiveDatasetManager(
            project_id='modified_project',
            dataset_name='modified_dataset',
            internal_tables=['table2'],
            external_tables={'table': 'table2'},
            extras={'param': 'param2'})

        # when
        job = standard_component.to_job(dependencies_override={'ds2': modified_dataset})

        # then
        job.execute(bigflow.JobContext.make())

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_not_mutate_dependency_configuration_when_redefining_dependencies(self, create_dataset_manager_mock):
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, {'dependency_kwargs': kwargs})

        # given
        default_dataset = InteractiveDatasetManager(
            project_id='default_project',
            dataset_name='default_dataset')

        @interactive_component(ds=default_dataset)
        def standard_component(ds):
            # then
            self.assertEqual(ds._dataset_manager['dependency_kwargs']['project_id'], 'default_project')
            self.assertEqual(ds._dataset_manager['dependency_kwargs']['dataset_name'], 'default_dataset')

        modified_dataset = InteractiveDatasetManager(
            project_id='modified_project',
            dataset_name='modified_dataset')

        # when
        standard_component.to_job(dependencies_override={'ds': modified_dataset})
        standard_component.to_job().execute(bigflow.JobContext.make())

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_wrap_each_dependency_into_operation_level_dataset_manager(self, create_dataset_manager_mock):
        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, {'dependency_kwargs': kwargs})
        dataset1 = InteractiveDatasetManager(project_id='project1', dataset_name='someds')

        @interactive_component(ds1=dataset1)
        def fake_component(ds1):
            # then
            self.assertEqual(ds1._peek, None)
            self.assertEqual(ds1._operation_name, None)
            self.assertEqual(ds1._peek_limit, None)

        # when
        fake_component.to_job().execute(bigflow.JobContext.make(runtime_str='2019-01-01'))


class InteractiveComponentRunTestCase(TestCase):

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_wrap_each_dependency_into_operation_level_dataset_manager(self, create_dataset_manager_mock):
        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, {'dependency_kwargs': kwargs})
        dataset1 = InteractiveDatasetManager(project_id='project1', dataset_name='someds')

        @interactive_component(ds1=dataset1)
        def fake_component(ds1):
            # then 1
            self.assertEqual(ds1._peek, None)
            self.assertEqual(ds1._operation_name, None)
            self.assertEqual(ds1._peek_limit, None)

        # when 1
        fake_component.run()

        @interactive_component(ds1=dataset1)
        def fake_component(ds1):
            # then 2
            self.assertEqual(ds1._peek, None)
            self.assertEqual(ds1._operation_name, 'some_operation')
            self.assertEqual(ds1._peek_limit, None)

        # when 2
        fake_component.run('2019-01-01', operation_name='some_operation')

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_pass_configuration_to_job(self, create_dataset_manager_mock):
        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, {'dependency_kwargs': kwargs})
        dataset1 = InteractiveDatasetManager(
            project_id='project1',
            dataset_name='someds')

        @interactive_component(ds1=dataset1)
        def fake_component(ds1):
            # then
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['project_id'], 'project1')
            self.assertEqual(ds1._dataset_manager['dependency_kwargs']['dataset_name'], 'someds')

        # when
        fake_component.run()


class InteractiveComponentPeekTestCase(TestCase):

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_wrap_each_dependency_into_operation_level_dataset_manager(self, create_dataset_manager_mock):
        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, mock.MagicMock())
        dataset = InteractiveDatasetManager(project_id='project1', dataset_name='dataset1')

        @interactive_component(ds=dataset)
        def fake_component(ds):
            # then
            self.assertEqual(ds._peek, True)
            self.assertEqual(ds._operation_name, 'some_operation')
            self.assertEqual(ds._peek_limit, 100)
            ds.write_truncate('some_table', '''some sql''', operation_name='some_operation')

        # when
        fake_component.peek('2019-01-01', operation_name='some_operation', limit=100)

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_pass_configuration_to_job(self, create_dataset_manager_mock):
        # given
        def fake_dataset_manager(kwargs):
            result = mock.Mock()
            result.configure_mock(**{
                'dependency_kwargs.return_value': kwargs
            })
            return result

        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, fake_dataset_manager(kwargs))
        dataset = InteractiveDatasetManager(
            project_id='project1',
            dataset_name='dataset1')

        @interactive_component(ds=dataset)
        def fake_component(ds):
            # then
            self.assertEqual(ds._dataset_manager.dependency_kwargs()['project_id'], 'project1')
            self.assertEqual(ds._dataset_manager.dependency_kwargs()['dataset_name'], 'dataset1')
            ds.write_truncate('some_table', '''some sql''', operation_name='some_operation')

        # when
        fake_component.peek('2019-01-01', operation_name='some_operation')

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_return_peeked_operation_result(self, create_dataset_manager_mock):
        # given
        def fake_dataset_manager(kwargs):
            result = mock.Mock()
            result.configure_mock(**{
                'collect.return_value': 'peeked_value'
            })
            return result

        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, fake_dataset_manager(kwargs))
        dataset = InteractiveDatasetManager(project_id='project1', dataset_name='dataset1')

        @interactive_component(ds=dataset)
        def fake_component(ds):
            ds.write_truncate('some_table', '''some sql''', operation_name='some_operation')

        # when
        peek_result = fake_component.peek('2019-01-01', operation_name='some_operation', limit=100)

        # then
        self.assertEqual(peek_result, 'peeked_value')

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_not_allow_peeking_with_none_argument(self, create_dataset_manager_mock):
        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, mock.Mock())
        dataset = InteractiveDatasetManager(project_id='project1', dataset_name='dataset1')

        @interactive_component(ds=dataset)
        def fake_component(ds):
            pass

        # then
        with self.assertRaises(ValueError):
            # when
            fake_component.peek('2019-01-01', operation_name=None)

        # then
        with self.assertRaises(ValueError):
            # when
            fake_component.peek('2019-01-01', limit=None)

        # then
        with self.assertRaises(ValueError):
            # when
            fake_component.peek(None)

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_throw_error_when_operation_not_found(self, create_dataset_manager_mock):
        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, mock.Mock())
        dataset = InteractiveDatasetManager(project_id='project1', dataset_name='dataset1')

        @interactive_component(ds=dataset)
        def fake_component(ds):
            pass

        # then
        with self.assertRaises(ValueError):
            # when
            fake_component.peek('2019-01-01', operation_name='nonexistent')


class InteractiveDatasetManagerTestCase(TestCase):

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_create_component_for_specified_operation(self, create_dataset_manager_mock):
        # given
        dataset_manager_mock = mock.Mock()
        create_dataset_manager_mock.side_effect = lambda **kwargs: (None, dataset_manager_mock)

        dataset = InteractiveDatasetManager(project_id='project1', dataset_name='dataset1')
        write_truncate_component = dataset.write_truncate('table', 'some sql')
        write_append_component = dataset.write_append('table', 'some sql')
        write_tmp_component = dataset.write_tmp('table', 'some sql')
        collect_component = dataset.collect('some sql')
        collect_list_component = dataset.collect_list('some sql', True)
        dry_run_component = dataset.dry_run('some sql')
        load_table_from_dataframe_component = dataset.load_table_from_dataframe('table', 'df')
        create_table_from_schema_component = dataset.create_table_from_schema('table', [{}])
        insert_component = dataset.insert('table', [{}])
        delete_dataset_component = dataset.delete_dataset()

        # when
        write_truncate_component.run()
        write_append_component.run()
        write_tmp_component.run()
        collect_component.run()
        collect_list_component.run()
        dry_run_component.run()
        load_table_from_dataframe_component.run()
        create_table_from_schema_component.run()
        insert_component.run()
        delete_dataset_component.run()

        # then
        dataset_manager_mock.assert_has_calls([
            mock.call.write_truncate(
                table_name='table',
                sql='some sql',
                custom_run_datetime=None,
                partitioned=True),
            mock.call.write_append(
                table_name='table',
                sql='some sql',
                custom_run_datetime=None,
                partitioned=True),
            mock.call.write_tmp(
                table_name='table',
                sql='some sql',
                custom_run_datetime=None),
            mock.call.collect(
                sql='some sql',
                custom_run_datetime=None),
            mock.call.collect_list(
                sql='some sql',
                custom_run_datetime=None,
                record_as_dict=True),
            mock.call.dry_run(
                sql='some sql',
                custom_run_datetime=None),
            mock.call.load_table_from_dataframe(
                table_name='table',
                df='df',
                custom_run_datetime=None,
                partitioned=True),
            mock.call.create_table_from_schema(
                table_name='table',
                schema=[{}],
                table=None),
            mock.call.insert(
                table_name='table',
                records=[{}],
                partitioned=True),
            mock.call.remove_dataset(),
        ])


def raise_import_error():
    raise ImportError()


PARAMETER1 = '1'
PARAMETER2 = '2'


class LogSyntaxErrorTestCase(TestCase):

    @mock.patch('bigflow.bigquery.interactive.logger')
    def test_should_catch_bad_request_and_log_message(self, logger_mock: mock.Mock):
        # when
        self.bad_request_occurred()

        # then
        logger_mock.error.assert_called_once_with('Syntax error')

    def test_should_pass_all_parameters(self):
        # expect
        self._test_should_pass_all_parameters(PARAMETER1, PARAMETER2)

    @log_syntax_error
    def _test_should_pass_all_parameters(self, par1, par2):
        self.assertEqual(par1, PARAMETER1)
        self.assertEqual(par2, PARAMETER2)

    @log_syntax_error
    def bad_request_occurred(self):
        raise BadRequest('Syntax error')


"""A PEP 517 interface to build backend used to build "bigflow" projects.

Custom wrappers/hooks for a build process might be added here eventually.
"""

from setuptools.build_meta import *
from setuptools.build_meta import __all__



### EXPERIMENTAL

import tempfile
import shutil
import logging
import os.path
import contextlib
import typing
import os
import csv
from apache_beam.pvalue import PCollection

import pandas as pd

import apache_beam as beam
import apache_beam.io as beam_io

import apache_beam.dataframe as df
import apache_beam.dataframe.io as df_io
import apache_beam.dataframe.convert as df_convert


from apache_beam.io.filesystems import FileSystems


logger = logging.getLogger(__name__)


@contextlib.contextmanager
def download_file_to_localfs(path: str):
    """Downloads single file from external filesystem (gcs, s3 etc)"""

    logger.info("Download file %s to local fs", path)
    with FileSystems.open(path) as a, tempfile.TemporaryDirectory() as d:
        fn = os.path.join(d, path.split("/", 2)[-1])
        os.makedirs(os.path.dirname(fn))
        with open(fn, 'bw+') as f:
            shutil.copyfileobj(a, f)
        yield fn


@contextlib.contextmanager
def download_and_unpack_to_localfs(path: str):
    """Download and unpack archive from external filesystem (s3://, gs://, etc)."""

    with tempfile.TemporaryDirectory() as tmpdir:
        with download_file_to_localfs(path) as fn:
            shutil.unpack_archive(fn, tmpdir)
        yield tmpdir


def _to_dataframe(x):
    return pd.DataFrame(x)


@beam.ptransform_fn
def ReadCSVFilesPlain(
    p: beam.Pipeline,
    file_pattern: str,
    fieldnames: typing.List[str],
):
    return (p
        | "Read files csv files"      >> beam_io.ReadFromText(file_pattern=file_pattern, skip_header_lines=1)
        | "Chunk for parsing"         >> beam.BatchElements()
        | "Parse csv lines to dicts"  >> beam.FlatMap(lambda x: map(dict, csv.DictReader(x, fieldnames=fieldnames)))
        | "Chunk for processing"      >> beam.BatchElements()
        | "Convert chunks to DF"      >> beam.Map(_to_dataframe)
    )


@beam.ptransform_fn
def WritePandasToCSV(
    pcoll: PCollection[pd.DataFrame],
    file_path_prefix: str,
    **kwargs,
):
    return (pcoll
        | "Convert DFs to tuples"   >> beam.FlatMap(lambda df: map(list, df.values))
        | "Convert to csv lines"    >> beam.MapTuple(lambda *args: ",".join(map(str, args)))
        | "Write results to csv"    >> beam_io.WriteToText(file_path_prefix=file_path_prefix, **kwargs)
    )


@beam.ptransform_fn
def ReadCSVToPandas(
    p: beam.Pipeline,
    *args,
    **kwargs,
) -> PCollection[pd.DataFrame]:
    data = p | "Read CSV" >> df_io.read_csv(*args, **kwargs)
    return df_convert.to_pcollection(data, yield_elements='pandas')



import datetime
import bigflow
import freezegun

from collections import OrderedDict
from unittest import TestCase, mock

from google.cloud import logging_v2

from bigflow.workflow import JobContext, Workflow, Definition, InvalidJobGraph, WorkflowJob


class WorkflowTestCase(TestCase):
    def test_should_run_jobs(self):
        # given
        definition = [mock.Mock() for i in range(100)]
        workflow = Workflow(workflow_id='test_workflow', definition=definition)

        # when
        workflow.run(datetime.datetime(2019, 1, 1))

        # then
        for step in definition:
            step.assert_has_calls([
                mock.call.execute(JobContext.make(
                    runtime=datetime.datetime(2019, 1, 1),
                    runtime_str='2019-01-01 00:00:00',
                    workflow=workflow,
                )),
            ])

    def test_should_run_single_job(self):
        # given
        first_job = mock.Mock(spec_set=['run', 'id'])
        first_job.id = 'first job'
        first_job.run = mock.Mock()

        definition = [first_job] + [mock.Mock() for i in range(100)]
        workflow = Workflow(workflow_id='test_workflow', definition=definition)

        # when
        workflow.run_job('first job', '2020-01-01')

        # then
        for step in definition[1:]:
            step.assert_not_called()
        # and
        first_job.assert_has_calls([mock.call.run('2020-01-01')])

    def test_should_run_single_job_with_context(self):
        # given
        first_job = mock.Mock(spec_set=['execute', 'run', 'id'])
        first_job.id = 'first job'
        first_job.execute = mock.Mock()

        definition = [first_job] + [mock.Mock() for i in range(100)]
        workflow = Workflow(workflow_id='test_workflow', definition=definition)

        # when
        workflow.run_job('first job', '2020-01-01')

        # then
        for step in definition[1:]:
            step.assert_not_called()

        first_job.execute.assert_called_once()
        ((ctx,), _kwargs) = first_job.execute.call_args

        self.assertIs(ctx.workflow, workflow)
        self.assertEqual(ctx.runtime_str, "2020-01-01")
        self.assertEqual(ctx.runtime, datetime.datetime(2020, 1, 1))

    def test_should_run_single_classbased_job_oldapi(self):

        # given
        class FirstJob:
            id = 'first job'
            runtime = None

            def run(self, runtime):
                assert self.runtime is None
                self.runtime = runtime

        first_job = FirstJob()
        workflow = Workflow(workflow_id='test_workflow', definition=[first_job])

        # when
        workflow.run_job('first job', "2020-01-01")

        # then
        self.assertEqual(first_job.runtime, "2020-01-01")

    def test_should_run_single_classbased_job(self):
        # given
        class FirstJob(bigflow.Job):
            id = 'first job'
            context = None

            def execute(self, context: JobContext):
                assert self.context is None
                self.context = context

        first_job = FirstJob()
        workflow = Workflow(workflow_id='test_workflow', definition=[first_job])

        # when
        workflow.run_job('first job', datetime.datetime(2020, 1, 1))

        # then
        context: JobContext = first_job.context
        self.assertIsNotNone(context)
        self.assertEqual(context.runtime, datetime.datetime(2020, 1, 1))
        self.assertIs(context.workflow, workflow)

    @freezegun.freeze_time("2020-01-02 03:04:05")
    def test_should_auto_fill_runtime_field(self):
        # when
        jc = bigflow.JobContext.make()

        # then
        self.assertIsInstance(jc.runtime, datetime.datetime)
        self.assertEqual(jc.runtime, datetime.datetime.now())
        self.assertEqual(jc.runtime_str, "2020-01-02 03:04:05")

    @mock.patch.dict('os.environ', bf_env="my-env")
    def test_should_auto_capture_env(self):
        # when
        jc = bigflow.JobContext.make()

        # then
        self.assertEqual(jc.env, "my-env")

    def test_should_auto_fill_workflow_id(self):
        # given
        workflow = mock.Mock()
        workflow.workflow_id = "the_workflow"

        # when
        jc = bigflow.JobContext.make(workflow=workflow)

        # then
        self.assertIs(jc.workflow, workflow)
        self.assertEqual(jc.workflow_id, "the_workflow")

    def test_should_parse_string_runtime(self):
        for dt_str, expected in [
            ("2020-01-02", datetime.datetime(2020, 1, 2)),
            ("2020-01-02 10:15:20", datetime.datetime(2020, 1, 2, 10, 15, 20)),
        ]:
            jc = bigflow.JobContext.make(runtime=dt_str)
            self.assertEqual(jc.runtime, expected)

    def test_should_convert_date_to_datetime(self):
        # given
        dt = datetime.date(2020, 1, 2)

        # when
        jc = bigflow.JobContext.make(runtime=dt)

        # then
        self.assertIsInstance(jc.runtime, datetime.datetime)
        self.assertEqual(jc.runtime, datetime.datetime(2020, 1, 2, 0, 0, 0))
        self.assertEqual(jc.runtime, datetime.datetime(2020, 1, 2, 0, 0, 0))

    def test_should_have_id_and_schedule_interval(self):
        # given
        workflow = Workflow(
            workflow_id='test_workflow',
            definition=[],
            schedule_interval='@hourly')

        # expected
        self.assertEqual(workflow.schedule_interval, '@hourly')

    def test_should_throw_exception_when_circular_dependency_is_found(self):
        # given
        original_job = mock.Mock()
        job1, job2, job3, job4 = [WorkflowJob(original_job, i) for i in range(4)]

        # job1 --- job2
        #   |        |
        #    \       |
        #     \      |
        #      \     |
        #       \    |
        #        \   |
        #         \  |
        #         job3

        job_graph = {
            job1: (job2,),
            job2: (job3,),
            job3: (job1,)
        }

        # expected
        with self.assertRaises(InvalidJobGraph):
            Definition(job_graph)

        # given

        # job1        job4
        #  |          | |
        #  |          | |
        #  |          | |
        # job2 ------ job3

        job_graph = {
            job1: (job2,),
            job2: (job3,),
            job3: (job4,),
            job4: (job3,)
        }

        # expected
        with self.assertRaises(InvalidJobGraph):
            Definition(job_graph)

    def test_should_run_jobs_in_order_accordingly_to_graph_schema(self):
        # given
        original_job = mock.Mock()
        job1, job2, job3, job4, job5, job6, job7, job8, job9 = [WorkflowJob(original_job, i) for i in range(9)]
        job_graph = OrderedDict([
            (job1, (job5, job6)),
            (job2, (job6,)),
            (job3, (job6,)),
            (job4, (job7,)),
            (job6, (job8,)),
            (job7, (job8,)),
            (job5, (job9,))
        ])

        #  job1     job2  job3  job4
        #    |  \    |    /      |
        #    |   \   |   /       |
        #    |    \  |  /        |
        #    |     \ | /         |
        #  job5    job6        job7
        #    |        \         /
        #    |         \       /
        #    |          \     /
        #    |           \   /
        #    |            \ /
        #   job9         job8

        definition = Definition(job_graph)
        workflow = Workflow(workflow_id='test_workflow', definition=definition, schedule_interval='@hourly')

        # expected
        self.assertEqual(list(workflow._build_sequential_order()), [job1, job5, job9, job2, job3, job6, job4, job7, job8])

        # given
        job_graph = OrderedDict([
            (job1, (job5, job6, job7)),
            (job2, (job6,)),
            (job3, (job6,)),
            (job4, (job7,)),
            (job6, (job8,)),
            (job7, (job8,)),
            (job5, (job9,)),
            (job6, (job9,))
        ])

        #  job1     job2  job3  job4
        #    |  \    |    /      |
        #    |   \   |   /       |
        #    |    \  |  /        |
        #    |     \ | /         |
        #  job5    job6        job7
        #    |      / \         /
        #    |     /   \       /
        #    |    /     \     /
        #    |   /       \   /
        #    |  /         \ /
        #   job9         job8

        definition = Definition(job_graph)
        workflow = Workflow(workflow_id='test_workflow', definition=definition, schedule_interval='@hourly')

        # expected
        self.assertEqual(workflow._build_sequential_order(), [job1, job5, job2, job3, job6, job9, job4, job7, job8])


import os
from unittest import TestCase, mock
from pathlib import Path

import responses

from bigflow.dagbuilder import get_dags_output_dir, clear_dags_output_dir
from bigflow.deploy import deploy_dags_folder, get_vault_token, deploy_docker_image


class DeployTestCase(TestCase):

    def test_should_clear_GCS_DAGs_folder(self):

        # given
        workdir = os.path.dirname(__file__)

        # mocking
        gs_client_mock = mock.Mock()
        bucket_mock = mock.Mock()
        users_dag_blob = mock.Mock()
        dags_blob = mock.Mock()
        monitoring_dag_blob = mock.Mock()
        users_dag_blob.name = 'my_dag.py'
        dags_blob.name = 'dags/'
        monitoring_dag_blob.name = 'dags/airflow_monitoring.py'
        bucket_mock.list_blobs.return_value = [users_dag_blob, dags_blob, monitoring_dag_blob]
        gs_client_mock.bucket.return_value = bucket_mock

        # when
        deploy_dags_folder(dags_dir=workdir + '/.dags', dags_bucket='europe-west1-1-bucket', project_id='', clear_dags_folder=True,
                           auth_method='local_account', gs_client=gs_client_mock)

        # then
        gs_client_mock.bucket.assert_called_with('europe-west1-1-bucket')
        users_dag_blob.delete.assert_called_with()
        dags_blob.delete.assert_not_called()
        monitoring_dag_blob.delete.assert_not_called()

    def test_should_upload_files_from_DAGs_output_dir_to_GCS_DAGs_folder(self):

        # given
        workdir = os.path.dirname(__file__)

        clear_dags_output_dir(workdir)
        dags_dir = get_dags_output_dir(workdir)
        f1 = dags_dir / 'zonk_1.txt'
        f1.touch()
        f2 = dags_dir / 'zonk_2.txt'
        f2.touch()

        # mocking
        gs_client_mock = mock.Mock()
        bucket_mock = mock.Mock()

        f1_blob_mock = mock.Mock()
        f2_blob_mock = mock.Mock()

        def blobs(name):
            if name == 'dags/zonk_1.txt':
                return f1_blob_mock
            if name == 'dags/zonk_2.txt':
                return f2_blob_mock
            return None

        gs_client_mock.bucket.return_value = bucket_mock
        bucket_mock.blob.side_effect = blobs

        # when
        deploy_dags_folder(dags_dir=workdir + '/.dags', dags_bucket='europe-west1-1-bucket', project_id='', clear_dags_folder=False,
                           auth_method='local_account', gs_client=gs_client_mock)

        # then
        gs_client_mock.bucket.assert_called_with('europe-west1-1-bucket')
        f1_blob_mock.upload_from_filename.assert_called_with(f1.as_posix(), content_type='application/octet-stream')
        f2_blob_mock.upload_from_filename.assert_called_with(f2.as_posix(), content_type='application/octet-stream')

    @responses.activate
    def test_should_retrieve_token_from_vault(self):
        # given
        responses.add(responses.GET, 'https://example.com/v1/gcp/token', status=200, json={'data': {'token': 'token_value'}})
        vault_endpoint = 'https://example.com/v1/gcp/token'

        # when
        token = get_vault_token(vault_endpoint, 'secret')

        # then
        self.assertEqual(token, 'token_value')
        self.assertEqual(len(responses.calls), 1)
        self.assertEqual(responses.calls[0].request.url, 'https://example.com/v1/gcp/token')
        self.assertEqual(responses.calls[0].request.headers['X-Vault-Token'], 'secret')

    @responses.activate
    def test_should_raise_value_error_if_vault_problem_occurred_during_fetching_token(self):
        # given
        responses.add(responses.GET, 'https://example.com/v1/gcp/token', status=503)
        vault_endpoint = 'https://example.com/v1/gcp/token'

        # then
        with self.assertRaises(ValueError):
            # when
            get_vault_token(vault_endpoint, 'secret')
            self.assertEqual(len(responses.calls), 1)
            self.assertEqual(responses.calls[0].request.url, 'https://example.com/v1/gcp/token')
            self.assertEqual(responses.calls[0].request.headers['X-Vault-Token'], 'secret')

    @mock.patch('bigflow.deploy.decode_version_number_from_file_name')
    @mock.patch('bigflow.deploy.load_image_from_tar')
    @mock.patch('bigflow.deploy._deploy_image_loaded_to_local_registry')
    @mock.patch('bigflow.deploy.remove_docker_image_from_local_registry')
    def test_should_remove_image_from_local_registry_after_deploy(self,
                                                                  remove_docker_image_from_local_registry,
                                                                  _deploy_image_loaded_to_local_registry,
                                                                  load_image_from_tar,
                                                                  decode_version_number_from_file_name):
        # given
        decode_version_number_from_file_name.return_value = 'version123'
        load_image_from_tar.return_value = 'image_id'

        # when
        deploy_docker_image('image-version123.tar', 'docker_repository')

        # then
        decode_version_number_from_file_name.assert_called_with(Path('image-version123.tar'))
        load_image_from_tar.assert_called_with('image-version123.tar')
        _deploy_image_loaded_to_local_registry.assert_called_with('version123', 'docker_repository', 'image_id', 'local_account', None, None)
        remove_docker_image_from_local_registry.assert_called_with('docker_repository:version123')

    @mock.patch('bigflow.deploy.decode_version_number_from_file_name')
    @mock.patch('bigflow.deploy.load_image_from_tar')
    @mock.patch('bigflow.deploy.tag_image')
    @mock.patch('bigflow.deploy.remove_docker_image_from_local_registry')
    def test_should_remove_image_from_local_registry_after_failed_deploy(self, remove_docker_image_from_local_registry,
                                                                         tag_image, load_image_from_tar,
                                                                         decode_version_number_from_file_name):

        # given
        decode_version_number_from_file_name.return_value = 'version123'
        load_image_from_tar.return_value = 'image_id'

        # when
        with self.assertRaises(ValueError):
            deploy_docker_image('image-version123.tar', 'docker_repository', 'invalid_auth_method')

        # then
        load_image_from_tar.assert_called_with('image-version123.tar')
        remove_docker_image_from_local_registry.assert_called_with('docker_repository:version123')


import logging
import contextlib
import io
import sys
import json
import os

from unittest import TestCase, mock

from google.cloud import logging_v2

import bigflow.log


class LoggerTestCase(TestCase):

    def configure_mocked_logging(self, project_id, log_name, workflow_id=None):
        self.gcp_handler = mock.Mock()
        self.gcp_handler.level = logging.INFO

        with mock.patch('bigflow.log.create_gcp_log_handler', return_value=self.gcp_handler):
            bigflow.log._LOGGING_CONFIGURED = False
            bigflow.log.init_logging(
                workflow_id=workflow_id,
                config={
                    'gcp_project_id': project_id,
                    'log_name': log_name,
                },
            )

    def setUp(self):
        self.configure_mocked_logging('project-id', 'logger_name', 'workflow-id')
        self.test_logger = logging.getLogger('any.random.logger.name')
        self.root_logger = logging.getLogger('')

    def _clear_all_root_loggers(self):
        for h in logging.getLogger().handlers[:]:
            logging.getLogger().removeHandler(h)
            h.close()

    def tearDown(self):
        self._clear_all_root_loggers()
        bigflow.log._LOGGING_CONFIGURED = False

    def test_should_create_correct_logging_link(self):
        # when
        f = io.StringIO()
        with contextlib.redirect_stderr(f):
            # stderr handler is created only when no other handlers are registered
            self._clear_all_root_loggers()
            self.configure_mocked_logging('project-id', 'another_log_name', 'workflow_id')

        # then
        out = f.getvalue()
        self.assertIn("LOGS LINK", out)
        self.assertIn("https://console.cloud.google.com/logs/query;query=", out)
        self.assertIn("labels.workflow_id%3D%22workflow_id%22", out)


    def _assert_single_log_event(self, message_re, severity=None):
        self.assertEqual(1, self.gcp_handler.handle.call_count)
        le: logging.LogRecord = self.gcp_handler.handle.call_args_list[0][0][0]

        if severity:
            self.assertEqual(le.levelname, severity)
        if message_re:
            self.assertRegexpMatches(le.message, message_re)

        return le

    def test_should_log_unhandled_exception(self):

        # when
        try:
            raise ValueError("oh no... i'm dying")
        except Exception:
            sys.excepthook(*sys.exc_info()) # simulate uncaught exception

        # then
        le = self._assert_single_log_event(
            message_re="Uncaught exception: oh no... i\'m dying",
            severity='ERROR',
        )
        self.assertTrue(le.exc_info)

    def test_should_handle_warning(self):
        # when
        self.test_logger.warning("warning message")

        # then
        self._assert_single_log_event(
            message_re="warning message",
            severity='WARNING',
        )

    def test_should_handle_info(self):
        # when
        self.test_logger.info("info message")

        # then
        self._assert_single_log_event(
            message_re="info message",
            severity='INFO',
        )

    def test_should_handle_error(self):
        # when
        self.test_logger.error("error message")

        # then
        self._assert_single_log_event(
            message_re="error message",
            severity='ERROR',
        )

    def test_should_install_gcp_handler_when_logging_already_exists(self):

        # given
        self._clear_all_root_loggers()
        logging.basicConfig(level=logging.ERROR)

        # when
        self.configure_mocked_logging('project-id', 'logger_name', 'workflow_id')
        self.test_logger.info("message")

        # then
        self._assert_single_log_event(
            message_re="message",
            severity='INFO',
        )


    @mock.patch.dict('os.environ')
    @mock.patch.dict('sys.modules')
    @mock.patch('google.cloud.logging.Client')
    @mock.patch('google.cloud.logging.handlers.CloudLoggingHandler')
    def test_logging_should_autoinitialize_via_env_variables(
        self,
        client_mock,
        cloud_logging_handler_mock
    ):
        # given
        for m in list(sys.modules):
            if m.startswith("bigflow."):
                del sys.modules[m]
        del sys.modules['bigflow']

        self._clear_all_root_loggers()
        os.environ['bf_log_config'] = json.dumps({'log_level': "INFO", "gcp_project_id": "proj"})

        # when
        import bigflow

        # then
        self.assertEqual(logging.root.level, logging.INFO)



import sys
from unittest import TestCase
from unittest import mock
from unittest.mock import patch

import apache_beam as beam
from apache_beam import Pipeline
from apache_beam.options.pipeline_options import PipelineOptions, GoogleCloudOptions, SetupOptions, StandardOptions, WorkerOptions
from apache_beam.runners.portability.fn_api_runner.fn_runner import RunnerResult
from apache_beam.testing.test_pipeline import TestPipeline
from collections import Counter
from bigflow import JobContext

from bigflow.dataflow import BeamJob
from bigflow.workflow import DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS, DEFAULT_PIPELINE_LEVEL_EXECUTION_TIMEOUT_SHIFT_IN_SECONDS


class CountWordsDriver:

    counter = Counter()

    def __init__(self):
        self.context = None
        self.pipeline = None
        self.counter.clear()

    @staticmethod
    def count(word, counts):
        CountWordsDriver.counter[word] += sum(counts)

    def run_ex(self, pipeline, context, words_to_count, words_to_filter):
        self.pipeline = pipeline
        self.context = context
        return (
            pipeline
            | 'LoadingWordsInput' >> beam.Create(words_to_count)
            | 'FilterWords'       >> beam.Filter(lambda w: w in words_to_filter)
            | 'MapToCount'        >> beam.Map(lambda w: (w, 1))
            | 'GroupWords'        >> beam.GroupByKey()
            | 'Count'             >> beam.MapTuple(self.count)
        )

    def run(self, pipeline: Pipeline, context: JobContext, driver_arguments: dict):
        return self.run_ex(pipeline, context, driver_arguments['words_to_count'], driver_arguments['words_to_filter'])

    def nope(self, pipeline, context):
        self.pipeline = pipeline
        self.context = context

    def nope_driverargs(self, pipeline, context, driver_arguments: dict):
        self.pipeline = pipeline
        self.context = context


class BeamJobTestCase(TestCase):

    def addMock(self, m):
        self.addCleanup(m.stop)
        return m.start()

    def setUp(self):
        super().setUp()

        # Beam tries to parse cmdargs eagerly - it breaks external test launchers
        self.addMock(mock.patch('sys.argv', ["python"]))
        self.addMock(mock.patch('bigflow.build.reflect.locate_project_path'))

    @patch.object(RunnerResult, 'state', new_callable=mock.PropertyMock)
    def test_should_run_beam_job(
        self,
        state_mock,
    ):
        # given
        state_mock.return_value = "DONE"

        driver = CountWordsDriver()
        job = BeamJob(
            id='count_words',
            entry_point=driver.run,
            entry_point_arguments={
                'words_to_filter': ['valid', 'word'],
                'words_to_count': ['trash', 'valid', 'word', 'valid']
            },
            test_pipeline=self._test_pipeline_with_label('count_words'))

        # when
        job.execute(JobContext.make())

        # then executes the job with the arguments
        self.assertEqual(driver.counter, {'valid': 2, 'word': 1})

        # and passes the context
        self.assertIsNotNone(driver.context)
        self.assertTrue(isinstance(driver.context, JobContext))

        # and labels the job
        self.assertEqual(
            driver.pipeline._options.get_all_options()['labels'],
            ['workflow_id=count_words'])

        # and sets default value for execution_timeout_sec
        self.assertEqual(job.execution_timeout_sec, DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS)

    @patch.object(RunnerResult, 'state', new_callable=mock.PropertyMock)
    def test_should_run_new_entry_point(
        self,
        state_mock,
    ):
        # given
        state_mock.return_value = "DONE"

        driver = CountWordsDriver()
        job = BeamJob(
            id='count_words',
            entry_point=driver.run_ex,
            entry_point_args=(
                ['trash', 'valid', 'word', 'valid'],
            ),
            entry_point_kwargs={
                'words_to_filter': ['valid', 'word'],
            },
            test_pipeline=self._test_pipeline_with_label('count_words'))

        # when
        job.execute(JobContext.make())

        # then executes the job with the arguments
        self.assertEqual(driver.counter, {'valid': 2, 'word': 1})

        # and passes the context
        self.assertIsNotNone(driver.context)
        self.assertTrue(isinstance(driver.context, JobContext))

        # and labels the job
        self.assertEqual(
            driver.pipeline._options.get_all_options()['labels'],
            ['workflow_id=count_words'])

    @patch.object(RunnerResult, 'state', new_callable=mock.PropertyMock)
    def test_should_run_old_entry_point_withoutargs(
        self,
        state_mock,
    ):
        # given
        state_mock.return_value = "DONE"

        driver = CountWordsDriver()
        job = BeamJob(
            id='count_words',
            entry_point=driver.nope_driverargs,
            test_pipeline=self._test_pipeline_with_label('count_words'))

        # when
        job.execute(JobContext.make())

        # then
        self.assertTrue(driver.context, "Driver was called")

    @patch.object(RunnerResult, 'state', new_callable=mock.PropertyMock)
    @patch.object(RunnerResult, 'cancel')
    @patch.object(RunnerResult, 'wait_until_finish')
    def test_should_run_beam_job_with_timeout_with_cancel(
        self,
        wait_until_finish_mock,
        cancel_mock,
        state_mock,
    ):
        # given
        wait_until_finish_mock.return_value = 'DONE'
        state_mock.return_value = 'RUNNING'

        driver = CountWordsDriver()

        job = BeamJob(
            id='count_words',
            entry_point=driver.nope,
            test_pipeline=self._test_pipeline_with_label('count_words'),
            execution_timeout_sec=600)

        # then
        with self.assertRaises(RuntimeError) as e:
            job.execute(JobContext.make())

        # then
        self.assertEqual(cancel_mock.call_count, 1)
        wait_until_finish_mock.assert_called_with((600 - DEFAULT_PIPELINE_LEVEL_EXECUTION_TIMEOUT_SHIFT_IN_SECONDS) * 1000)

    @patch.object(RunnerResult, 'state', new_callable=mock.PropertyMock)
    @patch.object(RunnerResult, 'cancel')
    @patch.object(RunnerResult, 'wait_until_finish')
    def test_should_run_beam_job_with_timeout_without_cancel(
        self,
        wait_until_finish_mock,
        cancel_mock,
        state_mock,
    ):
        # given
        wait_until_finish_mock.return_value = 'DONE'
        state_mock.return_value = 'DONE'

        driver = CountWordsDriver()
        job = BeamJob(
            id='count_words',
            entry_point=driver.nope,
            test_pipeline=self._test_pipeline_with_label('count_words'),
            execution_timeout_sec=600)

        # when
        job.execute(JobContext.make())

        # then
        self.assertEqual(cancel_mock.call_count, 0)
        wait_until_finish_mock.assert_called_with((600 - DEFAULT_PIPELINE_LEVEL_EXECUTION_TIMEOUT_SHIFT_IN_SECONDS) * 1000)

    def test_should_throw_if_wait_until_finish_set_to_false_and_execution_timeout_passed(
        self,
    ):
        # given
        with self.assertRaises(ValueError):
            driver = CountWordsDriver()
            job = BeamJob(
                id='count_words',
                entry_point=driver.nope,
                test_pipeline=self._test_pipeline_with_label('count_words'),
                execution_timeout_sec=1,
                wait_until_finish=False)

    @patch('bigflow.dataflow.Pipeline')
    def test_should_create_pipeline_from_pipeline_options(
        self,
        _create_pipeline_mock: mock.Mock,
    ):
        # given
        _create_pipeline_mock.return_value.run.return_value = RunnerResult('DONE', None)

        driver = CountWordsDriver()

        options = PipelineOptions()
        options.view_as(StandardOptions).runner = 'DataflowRunner'
        options.view_as(GoogleCloudOptions).project = 'gcp_project_id'
        options.view_as(GoogleCloudOptions).job_name = 'beam-wordcount-uuid'
        options.view_as(GoogleCloudOptions).staging_location = "gs://staging_location"
        options.view_as(GoogleCloudOptions).temp_location = "gs://temp_location"
        options.view_as(GoogleCloudOptions).region = 'region'
        options.view_as(GoogleCloudOptions).service_account_email = 'service-account'
        options.view_as(WorkerOptions).machine_type = 'n2-standard-8'
        options.view_as(WorkerOptions).max_num_workers = 2
        options.view_as(WorkerOptions).autoscaling_algorithm = 'THROUGHPUT_BASED'
        options.view_as(WorkerOptions).network = 'default'
        options.view_as(WorkerOptions).use_public_ips = False
        options.view_as(StandardOptions).runner = 'DataflowRunner'

        options.view_as(SetupOptions).setup_file = "/path/to/setup.py"

        job = BeamJob(
            id='count_words',
            entry_point=driver.nope,
            pipeline_options=options,
            execution_timeout_sec=10,
        )

        # when
        job.execute(JobContext.make())

        # then
        options.get_all_options()
        self.assertDictEqual(
            options.get_all_options(),
            _create_pipeline_mock.call_args[1]['options'].get_all_options(),
        )


    @patch('bigflow.dataflow.Pipeline')
    def test_should_create_pipeline_from_pipeline_options_when_argument_passed_to_sys_argv(
        self,
        _create_pipeline_mock: mock.Mock,
    ):
        # given
        sys.argv.append('build') # it simulates situation when we call `bf build`and build is passed to argv
        _create_pipeline_mock.return_value.run.return_value = RunnerResult('DONE', None)
        driver = CountWordsDriver()

        options = PipelineOptions()
        options.view_as(StandardOptions).runner = 'DataflowRunner'
        options.view_as(GoogleCloudOptions).project = 'gcp_project_id'
        options.view_as(GoogleCloudOptions).job_name = 'beam-wordcount-uuid'
        options.view_as(GoogleCloudOptions).staging_location = "gs://staging_location"
        options.view_as(GoogleCloudOptions).temp_location = "gs://temp_location"
        options.view_as(GoogleCloudOptions).region = 'region'
        options.view_as(GoogleCloudOptions).service_account_email = 'service-account'
        options.view_as(WorkerOptions).machine_type = 'n2-standard-8'
        options.view_as(WorkerOptions).max_num_workers = 2
        options.view_as(WorkerOptions).autoscaling_algorithm = 'THROUGHPUT_BASED'
        options.view_as(WorkerOptions).network = 'default'
        options.view_as(WorkerOptions).use_public_ips = False
        options.view_as(StandardOptions).runner = 'DataflowRunner'

        options.view_as(SetupOptions).setup_file = "/path/to/setup.py"
        job = BeamJob(
            id='count_words',
            entry_point=driver.nope,
            pipeline_options=options,
            execution_timeout_sec=10,
        )

        # when
        job.execute(JobContext.make())

        # then
        options.get_all_options()
        self.assertDictEqual(
            options.get_all_options(),
            _create_pipeline_mock.call_args[1]['options'].get_all_options(),
        )

    def test_should_throw_if_pipeline_options_and_pipeline_both_not_provided(
        self,
    ):
        with self.assertRaises(ValueError):
            driver = CountWordsDriver()
            BeamJob(
                id='count_words',
                entry_point=driver.nope,
                execution_timeout_sec=1)

    def test_should_throw_if_pipeline_options_and_pipeline_both_provided(
        self,
    ):
        with self.assertRaises(ValueError):
            driver = CountWordsDriver()
            BeamJob(
                id='count_words',
                entry_point=driver.nope,
                test_pipeline=self._test_pipeline_with_label('count_words'),
                execution_timeout_sec=1,
                pipeline_options=PipelineOptions())

    def _test_pipeline_with_label(self, workflow_id):
        pipeline_options = PipelineOptions()
        google_cloud_options = pipeline_options.view_as(GoogleCloudOptions)
        google_cloud_options.labels = []
        google_cloud_options.labels.append(f'workflow_id={workflow_id}')
        return TestPipeline(options=google_cloud_options)

    @patch('bigflow.dataflow.Pipeline')
    def test_create_pipeline_dict_options(
        self,
        _create_pipeline_mock: mock.Mock,
    ):
        # given
        _create_pipeline_mock.return_value.run.return_value = RunnerResult('DONE', None)

        driver = CountWordsDriver()
        options = {
            'job_name': "custom-my-job",
            'labels': ["lable1", "label2"],
            'streaming': True,
            'project': "my-gcp-project",
            'machine_type': "n2-standard-16",
        }

        job = BeamJob(
            id='count_words',
            entry_point=driver.nope,
            pipeline_options=options,
            execution_timeout_sec=10,
        )

        # when
        job.execute(JobContext.make())

        # then
        options2 = _create_pipeline_mock.call_args[1]['options'].get_all_options(drop_default=True)

        self.assertTrue(options2.pop('setup_file'))
        self.assertEqual(options2.pop('runner'), 'DataflowRunner')
        self.assertDictEqual(options, options2)

    @mock.patch('bigflow.dataflow.Pipeline')
    @mock.patch('bigflow.build.reflect.get_project_spec')
    def test_add_docker_to_pipelineoptions(
        self,
        get_project_spec_mock: mock.Mock,
        _create_pipeline_mock: mock.Mock,
    ):
        # given
        _create_pipeline_mock.return_value.run.return_value = RunnerResult('DONE', None)
        get_project_spec_mock.return_value.version = "1.2.3"
        get_project_spec_mock.return_value.docker_repository = "my_repo"

        driver = CountWordsDriver()
        job = BeamJob(
            id='count_words',
            entry_point=driver.nope,
            pipeline_options={},
            use_docker_image=True,
        )

        # when
        job.execute(JobContext.make())

        # then
        options2 = _create_pipeline_mock.call_args[1]['options'].get_all_options(drop_default=True)

        self.assertIn('job_name', options2)
        self.assertEqual('my_repo:1.2.3', options2['worker_harness_container_image'])
        self.assertIn('use_runner_v2', options2['experiments'])

    @patch('bigflow.dataflow.Pipeline')
    def test_add_setuppy_onle_for_dataflow_runner(
        self,
        _create_pipeline_mock: mock.Mock,
    ):
        # given
        _create_pipeline_mock.return_value.run.return_value = RunnerResult('DONE', None)

        driver = CountWordsDriver()
        options = {
            'job_name': "custom-my-job",
            'runner': 'DirectRunner',
        }

        job = BeamJob(
            id='count_words',
            entry_point=driver.nope,
            pipeline_options=options,
        )

        # when
        job.execute(JobContext.make())

        # then
        options2 = _create_pipeline_mock.call_args[1]['options'].get_all_options(drop_default=True)

        self.assertNotIn('setup_file', options2)


import os
import logging

from pathlib import Path

import jinja2
import jinja2.nodes
import jinja2.ext

logger = logging.getLogger(__name__)


class SkipThisTemplate(Exception):
    "May be raised from templates when the whole file should be skipped."


class SkipTemplateTagExtension(jinja2.ext.Extension):

    tags = {'skip_file_when', 'skip_file_unless'}

    def parse(self, parser):
        token = next(parser.stream)
        lineno = token.lineno
        cond = parser.parse_expression()
        return jinja2.nodes.CallBlock(
            self.call_method(token.value, [cond], lineno=lineno),
            [], [], [], lineno=lineno,
        )

    def skip_file_when(self, cond, caller):
        if cond:
            raise SkipThisTemplate()
        return ""

    def skip_file_unless(self, cond, caller):
        return self.skip_file_when(not cond, caller)


def render_templates(
    dest: os.PathLike,
    templates: jinja2.BaseLoader,
    variables: dict,
):
    dest = Path(dest)
    logger.info("render templates into %s", dest)

    env = jinja2.Environment(
        loader=templates,
        extensions=[SkipTemplateTagExtension],
    )
    env.filters['repr'] = repr
    env.filters['str'] = str

    logger.debug("jinja2 globals: %r", variables)
    env.globals = variables

    for template_name in env.list_templates():
        logger.debug("render template %r", template_name)

        name_expr = template_name
        if "." in template_name:
            base, ext = template_name.rsplit(".", 1)
            if ext in {'jinja', 'j2'}:
                name_expr = base

        name_template = env.from_string(name_expr)
        body_template = env.get_template(template_name)

        try:
            file_name = name_template.render()
            body = body_template.render()
        except SkipThisTemplate:
            logger.debug("skip template %r", template_name)
            continue

        dst_path = dest / file_name

        if not dst_path.parent.exists():
            dst_path.parent.mkdir(parents=True)

        try:
            dst_path.unlink()
        except FileNotFoundError:
            pass

        logger.info("create file %s", dst_path)
        dst_path.write_text(body)


def render_builtin_templates(dest: os.PathLike, template_dir: str, variables: dict):
    loader = jinja2.PackageLoader('bigflow.scaffold', f"templates/{template_dir}")
    logger.debug("load jinja2 templates from %s", loader)
    render_templates(dest, loader, variables)



from unittest import TestCase
from bigflow.bigquery.dataset_manager import handle_key_error, get_partition_from_run_datetime_or_none
from bigflow.bigquery.dataset_manager import AliasNotFoundError


class HandleKeyErrorTestCase(TestCase):
    def test_should_catch_key_error_and_raise_variable_not_found_error(self):
        # expect
        with self.assertRaises(AliasNotFoundError):
            self.raise_key_error()

    @handle_key_error
    def raise_key_error(self):
        '{missing_key}'.format(meh='bla')


class GetPartitionFromRunDatetimeOrNoneTestCase(TestCase):
    def test_should_get_hourly_partition(self):
        # expect
        self.assertEqual(get_partition_from_run_datetime_or_none("2021-01-01 00:00:00"), "2021010100")

    def test_should_get_daily_partition(self):
        # expect
        self.assertEqual(get_partition_from_run_datetime_or_none("2021-01-01"), "20210101")



import subprocess
import time
from subprocess import PIPE

from test import mixins

class _BaseTestCase(
    mixins.BfCliInteractionMixin,
    mixins.PrototypedDirMixin,
):
    pass


class TestMigrateFromV10(_BaseTestCase):
    proto_dir = "bf-projects/bf_simple_v10"

    def test_automigrate_to_v11_two(self):
        """Launch 'bigflow project-version' command from 'v1.0' project."""

        # when
        p = self.bigflow_spawn(["project-version"], timeout=10)

        # then
        p.expect(r"project_setup.py.*renamed.*setup.py")
        p.expect(r"Do you wan to run migration now\?")
        p.sendline("YES")

        p.read()  # ignore the rest of output
        self.assertFalse(p.wait())

        self.assertFileNotExists("project_setup.py")
        self.assertFileExists("pyproject.toml")
        self.assertFileExists("setup.py")


class TestMigrateFromV11(_BaseTestCase):
    proto_dir = "bf-projects/bf_simple_v11"

    def test_dont_migrate_when_already_v11(self):
        # when
        cp = self.bigflow_run(["project-version"], check=False, capture_output=True)

        # then nothing should actually happen :)
        self.assertFalse(cp.returncode)



import difflib
import os
from pathlib import Path

import mock
from unittest import TestCase
from bigflow.bigquery.job import Job
from bigflow.dagbuilder import get_dags_output_dir, clear_dags_output_dir, generate_dag_file, secret_template
from bigflow.workflow import WorkflowJob, Workflow, Definition, get_timezone_offset_seconds, hourly_start_time


class DagBuilderTestCase(TestCase):

    def test_should_get_DAGs_output_dir(self):
        # given
        workdir = os.path.dirname(__file__)

        # when
        dags_dir = get_dags_output_dir(workdir)

        # then
        self.assertEqual(dags_dir.as_posix(), workdir + "/.dags")

    def test_should_clear_DAGs_output_dir(self):
        # given
        workdir = os.path.dirname(__file__)
        dags_dir = get_dags_output_dir(workdir)
        f = dags_dir / 'zonk.txt'
        f.touch()

        # when
        clear_dags_output_dir(workdir)

        # then
        self.assertFalse(f.exists())

    @mock.patch('bigflow.workflow.get_timezone_offset_seconds')
    def test_should_generate_DAG_file_from_workflow_with_hourly_scheduling(self, get_timezone_offset_seconds_mock):
        # given
        workdir = os.path.dirname(__file__)
        get_timezone_offset_seconds_mock.return_value = 2 * 3600
        docker_repository = 'eu.gcr.io/my_docker_repository_project/my-project'

        # given
        job1 = Job(
            id='job1',
            component=mock.Mock(),
            retry_count=10,
            retry_pause_sec=20
        )
        job2 = Job(
            id='job2',
            component=mock.Mock(),
            retry_count=100,
            retry_pause_sec=200
        )
        job3 = Job(
            id='job3',
            component=mock.Mock(),
            retry_count=100,
            retry_pause_sec=200
        )
        w_job1 = WorkflowJob(job1, 1)
        w_job2 = WorkflowJob(job2, 2)
        w_job3 = WorkflowJob(job3, 3)
        graph = {
            w_job1: (w_job2, w_job3),
            w_job2: (w_job3,)
        }
        workflow = Workflow(
            workflow_id='my_workflow',
            definition=Definition(graph),
            start_time_factory=hourly_start_time,
            schedule_interval='@hourly')

        # when
        dag_file_path = generate_dag_file(workdir, docker_repository, workflow, '2020-07-02 10:00:00', '0.3.0', 'ca')

        # then
        self.assertEqual(dag_file_path, workdir + '/.dags/my_workflow__v0_3_0__2020_07_02_10_00_00_dag.py')

        dag_file_content = Path(dag_file_path).read_text()
        expected_dag_content = '''
import datetime
from airflow import DAG
from airflow.contrib.operators import kubernetes_pod_operator
from airflow.contrib.kubernetes import secret

default_args = {
            'owner': 'airflow',
            'depends_on_past': True,
            'start_date': datetime.datetime(2020, 7, 2, 8, 0),
            'email_on_failure': False,
            'email_on_retry': False,
            'execution_timeout': datetime.timedelta(seconds=10800),
}

dag = DAG(
    'my_workflow__v0_3_0__2020_07_02_10_00_00',
    default_args=default_args,
    max_active_runs=1,
    schedule_interval='@hourly'
)


tjob1 = kubernetes_pod_operator.KubernetesPodOperator(
    task_id='job1',
    name='job1',
    cmds=['bf'],
    arguments=['run', '--job', 'my_workflow.job1', '--runtime', '{{ execution_date.strftime("%Y-%m-%d %H:%M:%S") }}', '--project-package', 'ca', '--config', '{{var.value.env}}'],
    namespace='default',
    image='eu.gcr.io/my_docker_repository_project/my-project:0.3.0',
    is_delete_operator_pod=True,
    retries=10,
    retry_delay=datetime.timedelta(seconds=20),
    dag=dag,
    secrets=[],
    execution_timeout=datetime.timedelta(seconds=10800))


tjob2 = kubernetes_pod_operator.KubernetesPodOperator(
    task_id='job2',
    name='job2',
    cmds=['bf'],
    arguments=['run', '--job', 'my_workflow.job2', '--runtime', '{{ execution_date.strftime("%Y-%m-%d %H:%M:%S") }}', '--project-package', 'ca', '--config', '{{var.value.env}}'],
    namespace='default',
    image='eu.gcr.io/my_docker_repository_project/my-project:0.3.0',
    is_delete_operator_pod=True,
    retries=100,
    retry_delay=datetime.timedelta(seconds=200),
    dag=dag,
    secrets=[],
    execution_timeout=datetime.timedelta(seconds=10800))

tjob2.set_upstream(tjob1)

tjob3 = kubernetes_pod_operator.KubernetesPodOperator(
    task_id='job3',
    name='job3',
    cmds=['bf'],
    arguments=['run', '--job', 'my_workflow.job3', '--runtime', '{{ execution_date.strftime("%Y-%m-%d %H:%M:%S") }}', '--project-package', 'ca', '--config', '{{var.value.env}}'],
    namespace='default',
    image='eu.gcr.io/my_docker_repository_project/my-project:0.3.0',
    is_delete_operator_pod=True,
    retries=100,
    retry_delay=datetime.timedelta(seconds=200),
    dag=dag,
    secrets=[],
    execution_timeout=datetime.timedelta(seconds=10800))

tjob3.set_upstream(tjob2)
tjob3.set_upstream(tjob1)
'''

        self.assert_files_are_equal(expected_dag_content, dag_file_content)

    def test_should_pass_workflow_properties_to_airflow_dag(self):
        # given
        workdir = os.path.dirname(__file__)
        docker_repository = 'eu.gcr.io/my_docker_repository_project/my-project'

        # given
        job1 = Job(
            id='job1',
            component=mock.Mock(),
            retry_count=10,
            retry_pause_sec=20
        )
        w_job1 = WorkflowJob(job1, 1)
        graph = {
            w_job1: ()
        }
        workflow = Workflow(
            workflow_id='my_daily_workflow',
            definition=Definition(graph),
            depends_on_past=False,
            schedule_interval='@daily',
            secrets=['bf_secret_password', 'bf_secret_token'])

        # when
        dag_file_path = generate_dag_file(workdir, docker_repository, workflow, '2020-07-02', '0.3.0', 'ca')

        # then passes the depends_on_past parameter value
        self.assertEqual(dag_file_path, workdir + '/.dags/my_daily_workflow__v0_3_0__2020_07_02_00_00_00_dag.py')

        dag_file_content = Path(dag_file_path).read_text()
        expected_dag_content = '''
import datetime
from airflow import DAG
from airflow.contrib.operators import kubernetes_pod_operator
from airflow.contrib.kubernetes import secret

default_args = {
            'owner': 'airflow',
            'depends_on_past': False,
            'start_date': datetime.datetime(2020, 7, 1, 0, 0),
            'email_on_failure': False,
            'email_on_retry': False,
            'execution_timeout': datetime.timedelta(seconds=10800),
}

dag = DAG(
    'my_daily_workflow__v0_3_0__2020_07_02_00_00_00',
    default_args=default_args,
    max_active_runs=1,
    schedule_interval='@daily'
)


tjob1 = kubernetes_pod_operator.KubernetesPodOperator(
    task_id='job1',
    name='job1',
    cmds=['bf'],
    arguments=['run', '--job', 'my_daily_workflow.job1', '--runtime', '{{ execution_date.strftime("%Y-%m-%d %H:%M:%S") }}', '--project-package', 'ca', '--config', '{{var.value.env}}'],
    namespace='default',
    image='eu.gcr.io/my_docker_repository_project/my-project:0.3.0',
    is_delete_operator_pod=True,
    retries=10,
    retry_delay=datetime.timedelta(seconds=20),
    dag=dag,
    secrets=[secret.Secret(deploy_type='env', deploy_target='bf_secret_password', secret='bf-secret-password', key='bf_secret_password'), secret.Secret(deploy_type='env', deploy_target='bf_secret_token', secret='bf-secret-token', key='bf_secret_token')],
    execution_timeout=datetime.timedelta(seconds=10800))

'''
        self.assert_files_are_equal(expected_dag_content, dag_file_content)

    def test_should_generate_DAG_file_from_workflow_with_daily_scheduling(self):
        # given
        workdir = os.path.dirname(__file__)
        docker_repository = 'eu.gcr.io/my_docker_repository_project/my-project'

        # given
        job1 = Job(
            id='job1',
            component=mock.Mock(),
            retry_count=10,
            retry_pause_sec=20
        )
        w_job1 = WorkflowJob(job1, 1)
        graph = {
            w_job1: ()
        }
        workflow = Workflow(
            workflow_id='my_daily_workflow',
            definition=Definition(graph),
            schedule_interval='@daily')

        # when
        dag_file_path = generate_dag_file(workdir, docker_repository, workflow, '2020-07-02', '0.3.0', 'ca')

        # then
        self.assertEqual(dag_file_path, workdir + '/.dags/my_daily_workflow__v0_3_0__2020_07_02_00_00_00_dag.py')

        dag_file_content = Path(dag_file_path).read_text()
        expected_dag_content = '''
import datetime
from airflow import DAG
from airflow.contrib.operators import kubernetes_pod_operator
from airflow.contrib.kubernetes import secret

default_args = {
            'owner': 'airflow',
            'depends_on_past': True,
            'start_date': datetime.datetime(2020, 7, 1, 0, 0),
            'email_on_failure': False,
            'email_on_retry': False,
            'execution_timeout': datetime.timedelta(seconds=10800),
}

dag = DAG(
    'my_daily_workflow__v0_3_0__2020_07_02_00_00_00',
    default_args=default_args,
    max_active_runs=1,
    schedule_interval='@daily'
)


tjob1 = kubernetes_pod_operator.KubernetesPodOperator(
    task_id='job1',
    name='job1',
    cmds=['bf'],
    arguments=['run', '--job', 'my_daily_workflow.job1', '--runtime', '{{ execution_date.strftime("%Y-%m-%d %H:%M:%S") }}', '--project-package', 'ca', '--config', '{{var.value.env}}'],
    namespace='default',
    image='eu.gcr.io/my_docker_repository_project/my-project:0.3.0',
    is_delete_operator_pod=True,
    retries=10,
    retry_delay=datetime.timedelta(seconds=20),
    dag=dag,
    secrets=[],
    execution_timeout=datetime.timedelta(seconds=10800))

'''
        self.assert_files_are_equal(expected_dag_content, dag_file_content)

    def assert_files_are_equal(self, expected_dag_content, dag_file_content):
        if not expected_dag_content == dag_file_content:

            diff = list(difflib.Differ().compare(expected_dag_content.splitlines(keepends=True), dag_file_content.splitlines(keepends=True)))
            print('diff:')
            print(''.join(diff))

            raise ValueError('Files are not equal')

    def expected_start_date_shift(self) -> str :
        return get_timezone_offset_seconds()


import hashlib
import logging
import re
import subprocess
import time
import typing
import threading
import os

from pathlib import Path
from deprecated import deprecated
from datetime import datetime, timedelta


logger = logging.getLogger(__name__)


T = typing.TypeVar('T')


def public(
    *,
    alias_for: typing.Union[T, None] = None,
    class_alias: bool = False,
    deprecate_reason: typing.Optional[str] = None,
    deprecate_dropat: typing.Optional[str] = None,
):
    """Documentation decorator, used to mark function/class which should be considered as a public API.

    Only elements marked with this decorator may be treated as `stable API`.
    Any other elements may be deleted/changed without any warning or notice.

    When the decorator is applied to class all attributes/methods without `_` prefix are considered to be public.

    Optional argument `alias_for` indicates that marked element should be discarded and
    value of `alias_for` should be used instead.  Wrapped (discared) object still
    may provide type/signature hints for IDE/autocompletion.

    >> def _some_function(x, y, private_arg=None): return x, y, private_arg
    >> @public(alias_for=some_function)
    >> def some_function(x, y): ...

    >> sume_function is _some_function
    True

    Optional argument `class_alias` indicates that wrapped object is a class definition with
    a single base class.  This base class is used as a value for `alias_for`.
    It enables such pattern for defining class aliases:

    >> class _Origin: pass
    >> @public(class_alias=True)
    >> class Alias(_Origin):
    >>     pass
    >> _Origin is Alias
    True
    >> Alias.__name__
    "_Origin"

    """

    assert alias_for is None or not class_alias
    assert deprecate_reason or not deprecate_dropat

    def wrapper(f: T) -> T:
        if class_alias:
            assert isinstance(f, type)
            assert len(f.__bases__) == 1
            ff = f.__base__
        elif alias_for:
            ff = alias_for
        else:
            ff = f

        if f.__doc__ and not ff.__doc__:
            ff.__doc__ = f.__doc__  # alias is used - pick the docstring
        elif f is not ff and f.__doc__ and ff.__doc__:
            logging.warning("Both %r and %r have their docstrings", f, ff)

        if deprecate_reason or deprecate_dropat:
            return deprecated(reason=deprecate_reason)(ff)
        return ff

    return wrapper

@public(
    deprecate_reason="Use `str(x.absolute()) inliner instead",
    deprecate_dropat="2.0",
)
def resolve(path: Path):
    return str(path.absolute())


@public(
    deprecate_reason="Use `datetime.now().strftime('%Y-%m-%d %H:00:00')` instead.",
    deprecate_dropat="2.0",
)
def now(template: str = "%Y-%m-%d %H:00:00"):
    return datetime.now().strftime(template)


class _StreamOutputDumper(threading.Thread):
    "Dump stream to logger and collect results as a string."

    def __init__(
        self,
        process: subprocess.Popen,
        stream: typing.IO[str],
        callback: typing.Callable[[str], None],
    ):
        threading.Thread.__init__(self)
        self.process = process
        self.stream = stream
        self.callback = callback
        self._result_list = []
        self.start()

    def result(self):
        self.join()
        return "".join(self._result_list)

    def run(self):
        space_buffer = []

        while not self.stream.closed:
            try:
                line = self.stream.readline()
            except (ValueError, EOFError):
                return  # closed

            linee = line
            if linee.endswith("\n"):
                linee = linee[:-1]

            if linee.strip():
                # log line and prepend all buffered whitespaces
                space_buffer.append(linee)
                self.callback("".join(space_buffer))
                space_buffer.clear()
            else:
                # line contains only whitespaces
                space_buffer.append(line)
            self._result_list.append(line)


def run_process(
    args: typing.Union[str, typing.List],
    *,
    verbose=True,
    check=True,
    env_add=None,
    input=None,
    env=None,
    **kwargs,
):
    """Run external process, extends `subprocess.run`, returns 'stdout', may throw `subprocess.SubprocessError`.

    Arguments:
      args - string or list of strings/integers/paths etc, `None` is converted to empty string
      verbose - True if output of the command should be dumped into `logger.info`
      check - raise an exception if return code is not equal to 0
      input - send string to program stdin
      env - replace set of environment variables
      env_add - add new/overwrite environment variables (extends `os.environ`)

    """
    if isinstance(args, str):
        cmd = re.split(r"\s+", args)
    else:
        cmd = [
            str(x) if x is not None else ""
            for x in args
        ]

    logger.debug("cmd %r, kwargs %r", cmd, kwargs)
    if env_add:
        env = dict(env or os.environ)
        env.update(env_add)

    start = time.time()
    process = subprocess.Popen(
        cmd, text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE if input is not None else None,
        env=env,
        **kwargs,
    )

    stdout_dumper = _StreamOutputDumper(
        process, process.stdout, logger.info if verbose else logger.debug)
    stderr_dumper = _StreamOutputDumper(
        process, process.stderr, logger.error if verbose else logger.debug)

    if input:
        process.stdin.write(input)
        process.stdin.close()

    code = process.wait()

    process.stdout.close()
    process.stderr.close()
    stdout = stdout_dumper.result()
    stderr = stderr_dumper.result()

    if code and check:
        raise subprocess.CalledProcessError(
            process.returncode, cmd, output=stdout, stderr=stderr)

    duration = time.time() - start
    logger.debug("done in %s seconds, code %d", format(duration, ".2f"), process.returncode)

    return stdout


def generate_file_hash(fname: Path, algorithm: str = 'sha256') -> str:
    logger.debug("Calculate hash of %s", fname)
    h = hashlib.new(algorithm)
    h.update(fname.read_bytes())
    return algorithm + ":" + h.hexdigest()


def decode_version_number_from_file_name(file_path: Path):
    if file_path.suffix != '.tar':
        raise ValueError(f'*.tar file expected in {file_path.as_posix()}, got {file_path.suffix}')
    if not file_path.is_file():
        raise ValueError(f'File not found: {file_path.as_posix()}')

    split = file_path.stem.split('-', maxsplit=1)
    if not len(split) == 2:
        raise ValueError(f'Invalid file name pattern: {file_path.as_posix()}, expected: *-{{version}}.tar, for example: image-0.1.0.tar')
    return split[1]


def get_docker_image_id(tag):
    images = subprocess.getoutput(f"docker images -q {tag}")
    return images.split('\n')[0]


def build_docker_image_tag(docker_repository: str, package_version: str):
    package_version = package_version.replace("+", "-")  # fix "local version" separator
    return docker_repository + ':' + package_version


def remove_docker_image_from_local_registry(tag):
    print('Removing the image from the local registry')
    run_process(f"docker rmi {get_docker_image_id(tag)} --no-prune")


def as_timedelta(v: typing.Union[None, str, int, float, timedelta]) -> typing.Optional[timedelta]:
    if v is None:
        return None
    elif isinstance(v, timedelta):
        return v
    elif isinstance(v, (int, float)):
        return timedelta(seconds=v)
    elif v == "":
        return None
    else:
        return timedelta(seconds=float(v))


import os
import contextlib
import unittest
import unittest.mock
import pickle

import dill  # type: ignore

from bigflow.konfig import Konfig, resolve_konfig, dynamic, expand, fromenv


class TestKonfig(unittest.TestCase):

    def setUp(self) -> None:

        self.mocks = contextlib.ExitStack()
        self.addCleanup(self.mocks.close)

        # global mocks
        self.mocks.enter_context(unittest.mock.patch.dict('os.environ'))


    def test_simple_konfig(self):

        # given
        class C(Konfig):
            a = 1
            b = '2'

        # when
        config = C()

        # then
        self.assertEqual(config.a, 1)
        self.assertEqual(config.b, '2')
        self.assertEqual(dict(config), dict(a=1, b='2'))

    def test_override_props(self):

        # given
        class C(Konfig):
            a = 1
            b = '2'

        # when
        config = C(a=100, c=3)

        # then
        self.assertEqual(config.a, 100)
        self.assertEqual(config.b, '2')
        self.assertEqual(config.c, 3)

    def test_noclass_konfig(self):

        # when
        config = Konfig(a=1, b='2')

        # then
        self.assertEqual(config.a, 1)
        self.assertEqual(config.b, '2')
        self.assertEqual(dict(config), {'a': 1, 'b': '2'})

    def test_inheritance(self):

        # given
        class A(Konfig):
            x = "a"
            y = "a"

        class B(A):
            x = "b"

        class C(B):
            y = "config"

        class D(C):
            pass

        # when
        a = A()
        b = B()
        c = C()
        d = D()

        # then
        self.assertEqual(dict(a), dict(x="a", y="a"))
        self.assertEqual(dict(b), dict(x="b", y="a"))
        self.assertEqual(dict(c), dict(x="b", y="config"))
        self.assertEqual(dict(d), dict(x="b", y="config"))

    def test_inheritance_diamond(self):

        # given
        class A(Konfig):
            x = "A"
            y = "A"
            z = "A"

        class B(A):
            x = "B"
            z = "?"

        class C(A):
            y = "C"

        class D(C, B):
            z = "D"

        # when
        config = D()

        # then
        self.assertEqual(dict(config), dict(x="B", y="C", z="D"))

    def test_osenv_variables(self):
        # given
        os.environ.update(
            b='x',
            c='x',
            # d - miss
            e='1',
        )

        class C(Konfig):
            b = fromenv('b')
            c = fromenv('c', default="none")
            d = fromenv('d', default="none")
            e = fromenv('e', type=int)

        # when
        config = C()

        # then
        self.assertEqual(config.b, 'x')
        self.assertEqual(config.c, 'x')
        self.assertEqual(config.d, "none")
        self.assertEqual(config.e, 1)

    def test_osenv_missing(self):
        # given
        class C(Konfig):
            x = fromenv("missing")

        # expect
        with self.assertRaises(ValueError):
            C()

    def test_placeholder_expansion(self):
        # given
        class C(Konfig):
            a = expand("{e}")
            b = expand("pre{a}post")
            c = expand("{a}-{a}")
            d = expand("pre{{a}}post")
            e = expand("A")

        # when
        config = C()

        # then
        self.assertEqual(
            dict(config),
            dict(
                a="A",
                b="preApost",
                c="A-A",
                d="pre{a}post",
                e="A",
            ))

    def test_placeholder_expansion_inheritance(self):

        # given
        class A(Konfig):
            n = "A"
            a = "0"
            b = "2"
            c = "3"
            x = expand("{a}{b}{c}")
            y = expand("{x}{x}")
            z = expand("{y}")
            k = expand("{z}")

        class B(A):
            n = "B"
            a = "1"

        class C(B):
            n = "C"
            z = expand("{a}/{b}/{n}")

        # when
        config = C()

        # then
        self.assertEqual(
            dict(config),
            dict(
                n="C",
                a="1",
                b="2",
                c="3",
                x="123",
                y="123123",
                z="1/2/C",
                k="1/2/C",
            ))

    def test_resolve_configs(self):

        # given
        class A(Konfig): pass
        class B(A): pass
        class C(B): pass
        cm = {'a': A, 'b': B, 'c': C}

        # expect
        self.assertIs(resolve_konfig(cm, "a").__class__, A)
        self.assertIs(resolve_konfig(cm, "b").__class__, B)
        self.assertIs(resolve_konfig(cm, "c").__class__, C)

    def test_resolve_configs_fail(self):
        # given
        class A(Konfig): pass
        class B(A): pass

        # expect
        with self.assertRaises(ValueError):
            resolve_konfig({'b': B}, "A", lazy=False)

    def test_resolve_config_by_osenv(self):
        # given
        os.environ['bf_env'] = 'dev'

        class C(Konfig): pass
        class P(C): bb = 1
        class D(C): bb = 2

        # when
        config = resolve_konfig({'prod': P, 'dev': D})

        # then
        self.assertEqual(config.bb, 2)

    def test_resolve_config_to_default(self):
        # given
        class C(Konfig):
            pass
        class A(C):
            bb = 1
        class B(C):
            bb = 2

        # when
        config = resolve_konfig(dict(prod=A, dev=B), default='prod')

        # then
        self.assertEqual(config.bb, 1)

    def test_resolve_returns_lazy(self):
        # given
        getit = unittest.mock.Mock()

        class A(Konfig):
            x = dynamic(getit)

        # when
        a = resolve_konfig({'a': A}, 'a')

        # then
        getit.assert_not_called()
        getit.return_value = 123

        self.assertEqual(a.x, 123)
        getit.assert_called_once_with(a)

    class serialization_A(Konfig):
        a = "1"
        b = "2"
        x = fromenv('x')

    class serialization_B(serialization_A):
            a = "3"
            y = expand("{x}")

    def test_serialization(self):
        # given
        os.environ['x'] = "1"
        config = self.serialization_B(z="ok")

        # when
        s = pickle.dumps(config)
        config2 = pickle.loads(s)

        s = dill.dumps(config)
        config3 = dill.loads(s)

        # then
        self.assertEqual(dict(config), dict(config2))
        self.assertEqual(dict(config), dict(config3))

    def test_replace(self):

        # given
        class C(Konfig):
            a = 1
            b = '2'

        # when
        config = C().replace(a=123)

        # then
        self.assertEqual(config.a, 123)
        self.assertEqual(config.b, '2')



import bigflow

from inspect import getargspec

from bigflow.workflow import DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS
from .dataset_manager import create_dataset_manager

import logging
logger = logging.getLogger(__name__)

DEFAULT_RETRY_COUNT = 3
DEFAULT_RETRY_PAUSE_SEC = 60


class Job(bigflow.Job):

    def __init__(self,
                 component,
                 id=None,
                 retry_count=DEFAULT_RETRY_COUNT,
                 retry_pause_sec=DEFAULT_RETRY_PAUSE_SEC,
                 execution_timeout_sec=DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS,
                 **dependency_configuration):
        self.id = id or component.__name__
        logger.debug("Init bigquery Job with id %s", self.id)

        self.component = component
        self.dependency_configuration = dependency_configuration
        self.retry_count = retry_count
        self.retry_pause_sec = retry_pause_sec
        self.execution_timeout_sec = execution_timeout_sec

    def execute(self, context: bigflow.JobContext):
        logger.info("Execute job %s: %s", self.id, context)
        return self._run_component(self._build_dependencies(context.runtime_str))

    def _build_dependencies(self, runtime):
        deps = {
            dependency_name: self._build_dependency(
                dependency_config=self._find_config(dependency_name),
                runtime=runtime)
            for dependency_name in self._component_dependencies
        }
        logger.debug("Dependencies for %s are: %s", self.id, deps)
        return deps

    def _run_component(self, dependencies):
        return self.component(**dependencies)

    @property
    def _component_dependencies(self):
        cdeps = [dependency_name for dependency_name in getargspec(self._component).args]
        logger.debug("Component dependencies for %s are: %s", self.id, cdeps)
        return cdeps

    @property
    def _component(self):
        return self.component

    def _find_config(self, target_dependency_name):
        for dependency_name, config in self.dependency_configuration.items():
            if dependency_name == target_dependency_name:
                logger.debug("Found config for dependency %s: %s", target_dependency_name, config)
                return config
        raise ValueError("Can't find config for dependency: " + target_dependency_name)

    def _build_dependency(self, dependency_config, runtime):
        logger.debug("Build dataset manager for config %s", dependency_config)
        _, dataset_manager = create_dataset_manager(
            runtime=runtime,
            **dependency_config._as_dict())
        return dataset_manager


import os
import time
import inspect
import logging

from typing import List, Iterable
from pathlib import Path

from bigflow.commons import (
    public, resolve,
)


__all__ = [
    'find_all_resources',
    'read_requirements',
    'find_file',
    'get_resource_absolute_path',
    'find_setup',
    'create_file_if_not_exists',
    'create_setup_body',
    'find_or_create_setup_for_main_project_package',
    'resolve',
]

logger = logging.getLogger(__name__)


def find_all_resources(resources_dir: Path) -> Iterable[str]:
    for path in resources_dir.rglob('*'):
        if path.is_file():
            yield str(path.relative_to(resources_dir.parent))


@public(
    deprecate_reason="Use `bigflow.build.pip.read_requirements` instead",
)
def read_requirements(requirements_path: Path, recompile_check=True) -> List[str]:
    from bigflow.build.pip import read_requirements
    return read_requirements(requirements_path, recompile_check)


def find_file(file_name: str, search_start_file: Path, max_depth: int = 10) -> Path:
    '''
    Method tries to find the specified file in the hierarchy, starting the search from the search_start_file and going
    up the hierarchy.

    Example:
    your-project/
        project-package/
            __init__.py
            beam_process.py
        resources/
            requirements.txt
        setup.py

    Inside beam_process.py:
    options.view_as(SetupOptions).setup_file = find_file('setup.py', Path(__file__))
    '''
    for depth in range(1, max_depth + 1):
        current_node = search_start_file
        for i in range(1, depth + 1):
            current_node = current_node.parent
        path_to_check = current_node / file_name
        if os.path.exists(path_to_check):
            return path_to_check
    raise ValueError(f"Can't find the {file_name}")


@public()
def get_resource_absolute_path(resource_file_name: str, search_start_file: Path = None) -> Path:
    '''
    Method allows you to access a file from the resources directory.

    Example:
    your-project/
        project-package/
            __init__.py
            beam_process.py
        resources/
            requirements.txt
        setup.py

    Inside beam_process.py:
    options.view_as(SetupOptions).requirements_file = get_resource_absolute_path('requirements.txt')
    '''
    search_start_file = search_start_file or Path(inspect.getmodule(inspect.stack()[1][0]).__file__)
    resource_dir_path = find_file('resources', search_start_file)
    result = resource_dir_path / resource_file_name
    if not os.path.isfile(result):
        raise ValueError("Can't find the specified resource or resource is not a file.")
    return result


@public(
    deprecate_reason="""
        Don't use this method as it doesn't work when called from pip-installed package.
        Use `bigflow.build.reflect.materialize_setuppy` instead""",
)
def find_setup(search_start_file: Path, retries_left: int = 10, sleep_time: float = 5) -> Path:
    '''
    Method used for finding the setup.py for a Apache Beam process. Because the setup.py can be created on runtime,
    the find_setup method will retry search N times.

    Example:
    your-project/
        project-package/
            __init__.py
            beam_process.py
        setup.py

    Inside beam_process.py:
    options.view_as(SetupOptions).setup_file = find_setup(Path(__file__))
    '''
    try:
        return find_file('setup.py', search_start_file)
    except ValueError as e:
        if not retries_left:
            raise e
        time.sleep(sleep_time)
        return find_setup(search_start_file, retries_left - 1)


@public(
    deprecate_reason="""
        Call to this method can be replaced with inliner `(file_path.exists() or file_path.write_text(body)) and file_path`.
    """
)
def create_file_if_not_exists(file_path: Path, body: str) -> Path:
    if os.path.exists(file_path):
        return file_path
    with open(file_path, 'w+') as f:
        f.write(body)
    return file_path


@public(
    deprecate_reason="""
        This method doesn't work when called from pip-installed package.
        Use `bigflow.build.reflect.materialize_setuppy` instead"""
)
def create_setup_body(project_name: str) -> str:
    return f'''
import setuptools

setuptools.setup(
        name='{project_name}',
        version='0.1.0',
        packages=setuptools.find_packages(
                exclude=tuple(
                        p for p in setuptools.find_packages()
                        if not p.startswith('{project_name}'))))
'''


@public(
    deprecate_reason="""
        This method doesn't work when called from pip-installed package.
        Use `bigflow.build.reflect.materialize_setuppy` instead"""
)
def find_or_create_setup_for_main_project_package(project_name: str = None, search_start_file: Path = None) -> Path:
    from bigflow.build.reflect import materialize_setuppy
    return materialize_setuppy(project_name)



import datetime
from unittest import TestCase
import mock

import bigflow
from bigflow.bigquery.interactive import DatasetConfigInternal
from bigflow.bigquery.job import Job


class JobTestCase(TestCase):

    @mock.patch('bigflow.bigquery.job.create_dataset_manager')
    def test_should_run_bigflow_component(self, create_dataset_manager_mock):

        # given
        create_dataset_manager_mock.side_effect = lambda **kwargs: (kwargs, kwargs)

        def test_component(bigquery_dependency1, bigquery_dependency2):

            # then
            self.assertEqual(bigquery_dependency2, {
                'project_id': 'some-project-id-2',
                'dataset_name': 'some-dataset-2',
                'internal_tables': ['some_internal_table'],
                'external_tables': {'some_external_table': 'some.external.table'},
                'extras': {'extra_param': 'some-extra-param'},
                'runtime': '2019-01-01 00:00:00',
                'credentials': 'credentials',
                'location': 'EU'
            })

            # and
            self.assertEqual(bigquery_dependency1, {
                'project_id': 'some-project-id',
                'dataset_name': 'some-dataset',
                'internal_tables': ['some_internal_table'],
                'external_tables': {'some_external_table': 'some.external.table'},
                'extras': {'extra_param': 'some-extra-param'},
                'credentials': 'credentials',
                'runtime': '2019-01-01 00:00:00',
                'location': 'EU'
            })

        job = Job(component=test_component,
                  bigquery_dependency1=DatasetConfigInternal(
                      project_id='some-project-id',
                      dataset_name='some-dataset',
                      internal_tables=['some_internal_table'],
                      external_tables={'some_external_table': 'some.external.table'},
                      credentials='credentials',
                      extras={'extra_param': 'some-extra-param'}),
                  bigquery_dependency2=DatasetConfigInternal(
                      project_id='some-project-id-2',
                      dataset_name='some-dataset-2',
                      internal_tables=['some_internal_table'],
                      external_tables={'some_external_table': 'some.external.table'},
                      credentials='credentials',
                      extras={'extra_param': 'some-extra-param'}))

        # when
        job.execute(bigflow.JobContext.make(runtime=datetime.date(2019, 1, 1)))


from unittest import TestCase

from bigflow.bigquery.interactive import InteractiveComponent
from bigflow.bigquery.interactive import add_label


class FakeTable(object):
    def __init__(self, table_name):
        self.labels = {}
        self.table_name = table_name


class FakeClient(object):

    def get_table(self, table_name):
        return FakeTable(table_name)

    def update_table(self, table, fields):
        return table.table_name, table.labels


class FakeDataset(object):
    def __init__(self, project_id, dataset_name):
        self.project_id = project_id
        self.dataset_name = dataset_name

    @property
    def client(self):
        return FakeClient()

    @property
    def config(self):
        return {}


TEST_LABELS = {'my_label': 'test_label'}
TEST_PROJECT_ID = 'test_project'
TEST_DATASET_NAME = 'test_dataset'
TEST_TABLE_NAME = 'test_table'


class LabelsTestCase(TestCase):

    def test_should_return_interactive_component_when_ds_is_provided(self):
        # given
        dataset = FakeDataset(TEST_PROJECT_ID, TEST_DATASET_NAME)

        # when
        add_label_component = add_label(TEST_TABLE_NAME, TEST_LABELS, ds=dataset)

        # then
        self.assertIsInstance(add_label_component, InteractiveComponent)

        # and
        self.test_should_execute_update_table_with_proper_parameters(add_label_component._standard_component)

    def test_should_return_raw_component_when_ds_is_not_provided(self):
        # given
        add_label_component = add_label(TEST_TABLE_NAME, TEST_LABELS)

        # expect
        self.test_should_execute_update_table_with_proper_parameters(add_label_component)

    def test_should_execute_update_table_with_proper_parameters(self, add_label_component=None):
        # given
        add_label_component = add_label_component or add_label(TEST_TABLE_NAME, TEST_LABELS)
        dataset = FakeDataset(TEST_PROJECT_ID, TEST_DATASET_NAME)

        # expect
        self.assertEqual(add_label_component(dataset), (f'{TEST_PROJECT_ID}.{TEST_DATASET_NAME}.{TEST_TABLE_NAME}', TEST_LABELS))



"""Build-related developer tools and helpers.

Intented to be used by "CLI" and in "development environement".
"""

import sys
import logging
import tempfile
import pickle
import typing
import os
import os.path
import functools

from typing import Optional
from pathlib import Path

import bigflow.commons as bf_commons
from bigflow.commons import public


logger = logging.getLogger(__name__)

# Special "hidden" command-line argument for `setup.py`
# When passed as first argument triggers dumping projects parameters to temp file.
# See `bigflow.build.dist._maybe_dump_setup_params` for more details.
DUMP_PARAMS_SETUPPY_CMDARG = "__bigflow_dump_params"


@public()
def read_setuppy_args(
    path_to_setup: typing.Union[None, Path, str] = None,
    directory: typing.Union[None, Path, str] = None,
) -> dict:
    """Loads `setup.py`, returns all parameters of `bigflow.build.setup()` function.

    This function doesn't unpack 'embeeded sdist' archive when package is installed via pip.
    You could use `bigflow.build.materialize_setuppy` for such purposes, although it is not recommended"""

    assert directory is None or path_to_setup is None
    path_to_setup = path_to_setup or find_setuppy(directory)
    return _read_setuppy_args(path_to_setup)


@functools.lru_cache()
def _read_setuppy_args(path_to_setup: Path) -> dict:
    logger.info("Read project options from %s", path_to_setup)
    with tempfile.NamedTemporaryFile("r+b") as f:
        bf_commons.run_process(["python", path_to_setup, DUMP_PARAMS_SETUPPY_CMDARG, f.name], cwd=str(path_to_setup.parent))
        return pickle.load(f)


@public()
def find_project_dir(cwd: typing.Union[None, Path, str] = None) -> Path:
    """Find location of project root.
    Scan provided directory and its parents, searching for any `setup.py` or `pyproject.toml`.
    Scanning doesn't escape user home directory."""

    d = cwd = Path(cwd or Path.cwd()).resolve()
    while True:

        if (d / "pyproject.toml").exists():
            logger.debug("Found 'pyproject.toml' at %s", d)
            return d

        if (d / "setup.py").exists():
            logger.debug("Found 'setup.py' at %s", d)
            try:
                logger.debug("Check if %d is project root")
                read_setuppy_args(d)
            except Exception as e:
                logger.debug("Directory %s is not a project root", exc_info=e)
            else:
                return d

        if (d / "project_setup.py").exists():
            logger.warning("Found legacy `project_setup.py`, please rename it to `setup.py`")
            return d

        if (d / ".git").is_dir():
            logger.debug("Found '.git' at %s", d)
            return d

        if d == d.parent:
            logger.info("Stop project dir searching - reach root at %s", d)
            break  # root
        if d == Path.home():
            logger.info("Stop project root searching - reach home %d", d)
            break  # outside of home
        d = d.parent

    raise FileNotFoundError("Not found project root at %s", cwd)


@public(deprecate_reason="Use `find_project_dir`")
def find_setuppy(directory: typing.Union[None, Path, str] = None) -> Path:
    """Find location of project setup.py.

    Scan provided directory and its parents, searching for any `setup.py`
    Scanning doesn't escape user home directory."""

    project_dir = find_project_dir(directory)
    setup_py: Path = project_dir / "setup.py"
    prj_setup_py: Path = project_dir / "project_setup.py"

    if setup_py.exists():
        return setup_py
    elif prj_setup_py.exists():
        return prj_setup_py
    else:
        raise FileNotFoundError("Not found `setup.py` not `project_setup.py`")


def install_syspath(project_dir: Optional[Path] = None, chdir: bool = True):
    """Makes project files importable by 'bigflow' cli tool"""

    try:
        project_dir = project_dir or find_project_dir()
    except FileNotFoundError:
        logger.debug("Could not find `setup.py` - don't modify sys.path")
        return

    d = str(project_dir)
    if d not in sys.path:
        logger.debug("Add %r to `sys.path`", d)
        sys.path.insert(0, d)

    if chdir:
        os.chdir(d)


import abc
import collections
import typing
import warnings
import datetime as dt
import logging

import bigflow.configuration
from bigflow.commons import public


logger = logging.getLogger(__name__)


DEFAULT_SCHEDULE_INTERVAL = '@daily'

_RUNTIME_FORMATS = [
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%d",
]

DEFAULT_EXECUTION_TIMEOUT_IN_SECONDS = 10800  # 3 hours
DEFAULT_PIPELINE_LEVEL_EXECUTION_TIMEOUT_SHIFT_IN_SECONDS = 120  # 2 minutes


def get_timezone_offset_seconds() -> int:
    return dt.datetime.now().astimezone().tzinfo.utcoffset(None).seconds


@public()
def hourly_start_time(start_time: dt.datetime) -> dt.datetime:
    td = dt.timedelta(seconds=get_timezone_offset_seconds())
    return start_time.replace(microsecond=0) - td


@public()
def daily_start_time(start_time: dt.datetime) -> dt.datetime:
    td = dt.timedelta(hours=24)
    return start_time.replace(hour=0, minute=0, second=0, microsecond=0) - td


@public()
class JobContext(typing.NamedTuple):

    runtime: typing.Optional[dt.datetime]
    runtime_str: typing.Optional[str]

    workflow: typing.Optional['Workflow']
    workflow_id: typing.Optional[str]

    env: typing.Optional[str]
    # TODO: add unique 'workflow execution id' (for tracing/logging)

    @classmethod
    def make(
        cls,
        *,
        runtime: typing.Union[str, dt.datetime, None] = None,
        runtime_str: typing.Optional[str] = None,
        workflow: typing.Optional['Workflow'] = None,
        workflow_id: typing.Optional[str] = None,
        env: typing.Optional[str] = None,
    ):
        logger.debug("Build new JobContext...")

        if isinstance(runtime, str):
            # Parse 'runtime' when it is a string
            warnings.warn("Using `str` as `runtime` value is deprecated, please provide instance of `datetime`", DeprecationWarning)
            runtime_str = runtime_str or runtime
            runtime = _parse_runtime_str(runtime)
        elif isinstance(runtime, dt.date) and not isinstance(runtime, dt.datetime):
            logger.debug("Runtime has type `date`, convert to `datetime`")
            runtime = dt.datetime(runtime.year, runtime.month, runtime.day)
        elif runtime is None:
            # No 'runtime'  -  fallback to current date-time
            logger.info("No runtime specified - use current datetime %s", runtime)
            runtime = dt.datetime.now()

        if not runtime_str and runtime:
            runtime_str = runtime.strftime(_RUNTIME_FORMATS[0])

        if not env:
            logger.debug("No explicit env specified - try to capture it from env variables")
            env = bigflow.configuration.current_env()
            logger.debug("Env is %r", env)

        if workflow and workflow_id and workflow.workflow_id != workflow_id:
            raise ValueError(f"Explicitly specified `workflow_id` {workflow_id} doesn math to real {workflow.workflow_id}")

        if workflow and not workflow_id:
            workflow_id = workflow.workflow_id
            logger.debug("Workflow id is %s", workflow_id)
        elif workflow_id and not workflow:
            # TODO: Try to load/reconstruct workflow based on its id?
            pass

        jc = cls(
            runtime=runtime,
            runtime_str=runtime_str,
            workflow=workflow,
            workflow_id=workflow_id,
            env=env,
        )

        logger.debug("JobContext is %r", jc)
        return jc


@public()
class Job(abc.ABC):
    """Base abstract class for bigflow.Jobs.  It is recommended to inherit all your jobs from this class."""

    id: str
    retry_count: int = 3
    retry_pause_sec: int = 60
    execution_timeout_sec: int = 10800  # 3 hours

    def __init__(
        self,
        id=None,
        execution_timeout_sec=None,
        retry_count=None,
        retry_pause_sec=None,
    ):
        if id is not None:
            self.id = id

        if execution_timeout_sec is not None:
            self.execution_timeout_sec = execution_timeout_sec

        if retry_count is not None:
            self.retry_count = retry_count

        if retry_pause_sec is not None:
            self.retry_pause_sec = retry_pause_sec

    @abc.abstractmethod
    def execute(self, context: JobContext):
        raise NotImplementedError

    @public(deprecate_reason="Method `Job.run` is deprecated, use `Job.execute()` instead")
    def run(self, runtime):
        context = JobContext.make(runtime=runtime)
        return self.execute(context)


@public()
class Workflow(object):

    def __init__(
        self,
        workflow_id,
        definition,
        schedule_interval=DEFAULT_SCHEDULE_INTERVAL,
        start_time_factory: typing.Callable[[dt.datetime], dt.datetime] = daily_start_time,
        log_config: typing.Optional['bigflow.log.LogConfigDict'] = None,
        depends_on_past: bool = True,
        secrets: typing.Iterable[str] = ()
    ):
        self.definition = self._parse_definition(definition)
        self.schedule_interval = schedule_interval
        self.workflow_id = workflow_id
        self.start_time_factory = start_time_factory
        self.log_config = log_config
        self.depends_on_past = depends_on_past
        self.secrets = secrets

    @staticmethod
    def _execute_job(job, context):
        if not isinstance(job, Job):
            logger.debug("It is recommended to inherit your job %r from `bigflow.Job` class", job)
        if hasattr(job, 'execute'):
            job.execute(context)
        else:
            # fallback to old api
            warnings.warn("Old bigflow.Job api is used, please implement method `execute` (see bigflow.Job)")
            job.run(context.runtime_str)

    def _make_job_context(self, runtime):
        return JobContext.make(workflow=self, runtime=runtime)

    def run(self, runtime: typing.Union[dt.date, str, None] = None):
        context = self._make_job_context(runtime)
        for job in self._build_sequential_order():
            self._execute_job(job, context)

    def find_job(self, job_id) -> Job:
        for job_wrapper in self._build_sequential_order():
            if job_wrapper.job.id == job_id:
                return job_wrapper.job
        raise ValueError(f'Job {job_id} not found.')

    def run_job(self, job_id: str, runtime: typing.Union[dt.date, str, None] = None):
        context = self._make_job_context(runtime)
        self._execute_job(self.find_job(job_id), context)

    def _build_sequential_order(self):
        return self.definition._sequential_order()

    def _call_on_graph_nodes(self, consumer):
        self.definition._call_on_graph_nodes(consumer)

    def _parse_definition(self, definition):
        if isinstance(definition, list):
            return Definition(self._map_to_workflow_jobs(definition))
        if isinstance(definition, Definition):
            return definition
        raise ValueError("Invalid argument %s" % definition)

    @staticmethod
    def _map_to_workflow_jobs(job_list):
        return [WorkflowJob(job, i) for i, job in enumerate(job_list)]


class WorkflowJob(Job):

    def __init__(self, job, name):
        self.job = job
        self.name = name

    @property
    def id(self):
        return self.job.id

    @property
    def retry_count(self):
        return self.job.retry_count

    @property
    def retry_pause_sec(self):
        return self.job.retry_pause_sec

    def execute(self, context: JobContext):
        Workflow._execute_job(self.job, context)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name

    def __repr__(self):
        return "WorkflowJob{job=..., name=%s}" % self.name


@public()
class Definition:
    def __init__(self, jobs: dict):
        self.job_graph = self._build_graph(jobs)
        self.job_order_resolver = JobOrderResolver(self.job_graph)

    def _sequential_order(self):
        return self.job_order_resolver.find_sequential_run_order()

    def _call_on_graph_nodes(self, consumer):
        return self.job_order_resolver._call_on_graph_nodes(consumer)

    def _build_graph(self, jobs):
        if isinstance(jobs, list):
            job_graph = self._convert_list_to_graph(jobs)
        elif isinstance(jobs, dict):
            job_graph = {self._map_to_workflow_job(source_job): [self._map_to_workflow_job(tj) for tj in target_jobs]
                         for source_job, target_jobs in jobs.items()}
        else:
            raise ValueError("Job graph has to be dict or list")

        JobGraphValidator(job_graph).validate()
        return job_graph

    def _map_to_workflow_job(self, job):
        if not isinstance(job, WorkflowJob):
            job = WorkflowJob(job, job.id)
        return job

    @staticmethod
    def _convert_list_to_graph(job_list):
        graph_as_dict = collections.OrderedDict()
        if len(job_list) == 1:
            graph_as_dict[job_list[0]] = []
        else:
            for i in range(1, len(job_list)):
                graph_as_dict[job_list[i - 1]] = [job_list[i]]
        return graph_as_dict


class InvalidJobGraph(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __repr__(self):
        return self.msg


class JobGraphValidator:
    def __init__(self, job_graph):
        self.job_graph = job_graph

    def validate(self):
        self._validate_if_not_cyclic()

    def _validate_if_not_cyclic(self):
        visited = set()
        stack = set()
        for job in self.job_graph:
            self._validate_job(job, visited, stack)

    def _validate_job(self, job, visited, stack):
        if job in stack:
            raise InvalidJobGraph(f"Found cyclic dependency on job {job}")
        if job in visited:
            return

        visited.add(job)
        if not job in self.job_graph:
            return
        stack.add(job)
        for dep in self.job_graph[job]:
            self._validate_job(dep, visited, stack)
        stack.remove(job)


class JobOrderResolver:
    def __init__(self, job_graph):
        self.job_graph = job_graph
        self.parental_map = self._build_parental_map()

    def find_sequential_run_order(self):
        ordered_jobs = []

        def add_to_ordered_job(job, dependencies):
            ordered_jobs.append(job)

        self._call_on_graph_nodes(add_to_ordered_job)
        return ordered_jobs

    def _call_on_graph_nodes(self, consumer):
        visited = set()
        for job in self.parental_map:
            self._call_on_graph_node_helper(
                job, self.parental_map, visited, consumer)

    def _build_parental_map(self):
        visited = set()
        parental_map = collections.OrderedDict()
        for job in self.job_graph:
            self._fill_parental_map(job, parental_map, visited)
        return parental_map

    def _fill_parental_map(self, job, parental_map, visited):
        if job not in self.job_graph or job in visited:
            return
        visited.add(job)

        if job not in parental_map:
            parental_map[job] = []

        for dependency in self.job_graph[job]:
            if dependency not in parental_map:
                parental_map[dependency] = []
            parental_map[dependency].append(job)
            self._fill_parental_map(dependency, parental_map, visited)

    def _call_on_graph_node_helper(self, job, parental_map, visited, consumer):
        if job in visited:
            return
        visited.add(job)

        for parent in parental_map[job]:
            self._call_on_graph_node_helper(
                parent, parental_map, visited, consumer)
        consumer(job, parental_map[job])


def _parse_runtime_str(runtime: str):
    for format in _RUNTIME_FORMATS:
        try:
            return dt.datetime.strptime(runtime, format)
        except ValueError:
            pass
    raise ValueError("Unable to parse 'runtime' %r" % runtime)



import functools
import json
import logging
import re
import typing
import uuid
from pathlib import Path

# hidden BQ and pandas imports due to https://github.com/allegro/bigflow/issues/149


logger = logging.getLogger(__name__)


DEFAULT_REGION = 'europe-west1'
DEFAULT_MACHINE_TYPE = 'n1-standard-1'
DEFAULT_LOCATION = 'EU'


class AliasNotFoundError(ValueError):
    pass


def handle_key_error(method):
    logger.debug("Wrap %s with @handle_key_error", method)

    @functools.wraps(method)
    def decorated(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except KeyError as e:
            missing_variable = e.args[0]
            raise AliasNotFoundError(
                "'{missing_variable}' is missing in internal_tables or external_tables or extras.".format(
                    missing_variable=missing_variable))

    return decorated


def get_partition_from_run_datetime_or_none(run_datetime: str) -> str:
    """
    :param run_datetime: string run datetime in format YYYY-MM-DD HH:mm:ss or YYYY-MM-DD
    :return: string partition in format YYYYMMDDHH for run_datetime in format: YYYY-MM-DD HH:mm:ss or
     YYYYMMDD for run_datetime in format YYYY-MM-DD
    """
    if run_datetime:
        assert re.fullmatch(r"\d{4}-\d\d-\d\d( \d\d:\d\d:\d\d)?", run_datetime), "run_datetime in invalid format"
        return re.sub(r"\D", "", run_datetime)[:10]


class TemplatedDatasetManager(object):
    """
    Decorator that resolves table names/all kind of variables.
    """
    def __init__(self,
                 dataset_manager,
                 internal_tables,
                 external_tables,
                 extras,
                 run_datetime):
        self.dataset_manager = dataset_manager
        self.internal_tables = {t: self.create_full_table_id(t) for t in internal_tables}
        self.external_tables = external_tables
        self.extras = extras
        self.run_datetime = run_datetime

        logger.debug(
            "Wrap %s with TemplatedDatasetManager, internal_tables %s,"
            " external_tables %s, extras %s, run_datetime %s",
            dataset_manager,
            self.internal_tables,
            self.external_tables,
            self.extras,
            self.run_datetime,
        )

    def write_truncate(self, table_name, sql, custom_run_datetime=None):
        return self.write(self.dataset_manager.write_truncate, table_name, sql, custom_run_datetime)

    def write_append(self, table_name, sql, custom_run_datetime=None):
        return self.write(self.dataset_manager.write_append, table_name, sql, custom_run_datetime)

    def write_tmp(self, table_name, sql, custom_run_datetime=None):
        self.internal_tables[table_name] = self.create_table_id(table_name)
        return self.write(self.dataset_manager.write_tmp, table_name, sql, custom_run_datetime)

    @handle_key_error
    def write(self, write_callable, table_name, sql, custom_run_datetime=None):
        table_id = self.create_table_id(table_name)
        return write_callable(table_id, sql.format(**self.template_variables(custom_run_datetime)))

    def create_table_id(self, table_name):
        table_name_without_partition = table_name.split('$')[0]
        return table_name.replace(
            table_name_without_partition,
            self.create_full_table_id(table_name_without_partition))

    @handle_key_error
    def collect(self, sql, custom_run_datetime=None):
        return self.dataset_manager.collect(sql.format(**self.template_variables(custom_run_datetime)))

    @handle_key_error
    def collect_list(self, sql: str, custom_run_datetime: typing.Optional[str] = None, record_as_dict: bool = False):
        return self.dataset_manager.collect_list(
            sql.format(**self.template_variables(custom_run_datetime)), record_as_dict)

    def dry_run(self, sql, custom_run_datetime=None):
        return self.dataset_manager.dry_run(sql.format(**self.template_variables(custom_run_datetime)))

    def remove_dataset(self):
        return self.dataset_manager.remove_dataset()

    def load_table_from_dataframe(self, table_name, df):
        table_id = self.create_table_id(table_name)
        return self.dataset_manager.load_table_from_dataframe(table_id, df)

    def create_table(self, create_query):
        return self.dataset_manager.create_table(create_query)

    def create_full_table_id(self, table_name):
        return self.dataset_manager.dataset_id + '.' + table_name

    def table_exists(self, table_name):
        return self.dataset_manager.table_exists(table_name)

    def template_variables(self, custom_run_datetime=None):
        result = {}
        result.update(self.internal_tables)
        result.update(self.external_tables)
        result.update(self.extras)
        result['dt'] = custom_run_datetime or self.run_datetime
        return result

    def create_table_from_schema(
            self,
            table_name: str,
            schema: typing.Union[None, typing.List[dict], Path] = None,
            table=None):
        table_id = self.create_table_id(table_name)
        return self.dataset_manager.create_table_from_schema(table_id, schema, table)

    def insert(
            self,
            table_name: str,
            records: typing.Union[typing.List[dict], Path]):
        table_id = self.create_table_id(table_name)
        return self.dataset_manager.insert(table_id, records)


class PartitionedDatasetManager(object):
    """
    Interface available for user. Manages partitioning.
    Delegate rest of the tasks to TemplatedDatasetManager and DatasetManager.
    """
    def __init__(self, templated_dataset_manager: TemplatedDatasetManager, partition):
        self._dataset_manager = templated_dataset_manager
        self.partition = partition

    def write_truncate(self, table_name, sql, partitioned=True, custom_run_datetime=None):
        return self._write(
            self._dataset_manager.write_truncate,
            table_name,
            sql,
            partitioned,
            custom_run_datetime)

    def write_append(self, table_name, sql, partitioned=True, custom_run_datetime=None):
        return self._write(
            self._dataset_manager.write_append,
            table_name,
            sql,
            partitioned,
            custom_run_datetime)

    def write_tmp(self, table_name, sql, custom_run_datetime=None):
        return self._write(
            self._dataset_manager.write_tmp,
            table_name,
            sql,
            False,
            custom_run_datetime)

    def collect(self, sql, custom_run_datetime=None):
        return self._dataset_manager.collect(sql, custom_run_datetime)

    def collect_list(self, sql: str, custom_run_datetime: typing.Optional[str] = None, record_as_dict: bool = False):
        return self._dataset_manager.collect_list(sql, custom_run_datetime, record_as_dict)

    def dry_run(self, sql, custom_run_datetime=None):
        return self._dataset_manager.dry_run(sql, custom_run_datetime)

    def create_table(self, create_query):
        return self._dataset_manager.create_table(create_query)

    @property
    def runtime_str(self):
        return self._dataset_manager.run_datetime

    @property
    def extras(self):
        return self._dataset_manager.extras

    @property
    def client(self):
        return self._dataset_manager.dataset_manager.bigquery_client

    @property
    def project_id(self):
        return self._dataset_manager.dataset_manager.dataset.project

    @property
    def dataset_name(self):
        return self._dataset_manager.dataset_manager.dataset.dataset_id

    @property
    def internal_tables(self):
        return self._dataset_manager.internal_tables

    @property
    def external_tables(self):
        return self._dataset_manager.external_tables

    def remove_dataset(self):
        return self._dataset_manager.remove_dataset()

    def load_table_from_dataframe(self, table_name, df, partitioned=True, custom_run_datetime=None):
        table_id = self._create_table_id(custom_run_datetime, table_name, partitioned)
        return self._dataset_manager.load_table_from_dataframe(table_id, df)

    def create_table_from_schema(
            self,
            table_name: str,
            schema: typing.Optional[typing.Union[typing.List[dict], Path]] = None,
            table=None):
        return self._dataset_manager.create_table_from_schema(table_name, schema, table)

    def insert(
            self,
            table_name: str,
            records: typing.Union[typing.List[dict], Path],
            partitioned: bool = True,
            custom_run_datetime: typing.Optional[str] = None):
        table_id = self._create_table_id(custom_run_datetime, table_name, partitioned)
        return self._dataset_manager.insert(table_id, records)

    def _write(self, write_callable, table_name, sql, partitioned, custom_run_datetime=None):
        table_id = self._create_table_id(custom_run_datetime, table_name, partitioned)
        return write_callable(table_id, sql, custom_run_datetime)

    def _create_table_id(self, custom_run_datetime, table_name, partitioned):
        custom_partition = get_partition_from_run_datetime_or_none(custom_run_datetime)
        if partitioned:
            table_name = table_name + '${partition}'.format(partition=custom_partition or self.partition)
        return table_name

    def _table_exists(self, table_name):
        return self._dataset_manager.table_exists(table_name)


class DatasetManager(object):
    """
    Manages BigQuery IO operations.
    """
    def __init__(self,
                 bigquery_client,
                 dataset,
                 logger):
        from google.cloud import bigquery
        self.bigquery_client: bigquery.Client  = bigquery_client
        self.dataset = dataset
        self.dataset_id = dataset.full_dataset_id.replace(':', '.')
        self.logger = logger

    def write_tmp(self, table_id, sql):
        return self.write(table_id, sql, 'WRITE_TRUNCATE')

    def write(self, table_id, sql, mode):
        from google.cloud import bigquery
        self.logger.info('%s to %s', mode, table_id)
        job_config = bigquery.QueryJobConfig()
        job_config.use_legacy_sql = False
        job_config.allow_large_results = True

        job_config.destination = table_id
        job_config.write_disposition = mode

        job = self.bigquery_client.query(sql, job_config=job_config)
        return job.result()

    def write_truncate(self, table_id, sql):
        self.table_exists_or_error(table_id)
        return self.write(table_id, sql, 'WRITE_TRUNCATE')

    def write_append(self, table_id, sql):
        self.table_exists_or_error(table_id)
        return self.write(table_id, sql, 'WRITE_APPEND')

    def table_exists_or_error(self, table_id):
        table_name = table_id.split('$')[0].split('.')[2]
        if not self.table_exists(table_name):
            raise ValueError('Table {id} does not exist'.format(id=table_id))

    def create_table(self, create_query):
        from google.cloud import bigquery
        self.logger.info('CREATE TABLE: %s', create_query)
        job_config = bigquery.QueryJobConfig()
        job_config.use_legacy_sql = False
        job_config.default_dataset = self.dataset

        job = self.bigquery_client.query(
            create_query,
            job_config=job_config)
        return job.result()

    def collect(self, sql):
        return self._query(sql).to_dataframe()

    def collect_list(self, sql: str, record_as_dict: bool = False):
        result = list(self._query(sql).result())
        if record_as_dict:
            result = [dict(e) for e in result]
        return result

    def dry_run(self, sql):
        from google.cloud import bigquery
        job_config = bigquery.QueryJobConfig()
        job_config.dry_run = True
        query_job = self._query(sql, job_config=job_config)
        billed = self._convert_to_humanbytes(query_job.total_bytes_processed)
        return "This query will process {} and cost {}.".format(
          billed['size'],
          billed['cost'])

    def remove_dataset(self):
        return self.bigquery_client.delete_dataset(self.dataset, delete_contents=True, not_found_ok=True)

    def load_table_from_dataframe(self, table_id, df):
        return self.bigquery_client.load_table_from_dataframe(df, table_id).result()

    def table_exists(self, table_name):
        return self.bigquery_client.query('''
            SELECT count(*) as table_exists
            FROM `{dataset_id}.__TABLES__`
            WHERE table_id='{table_name}'
            '''.format(
                dataset_id=self.dataset_id,
                table_name=table_name)) \
            .result() \
            .to_dataframe()['table_exists'] \
            .iloc[0] > 0

    def create_table_from_schema(
            self,
            table_id: str,
            schema: typing.Union[typing.List[dict], Path, None] = None,
            table=None):
        from google.cloud.bigquery import Table, TimePartitioning

        if schema and table:
            raise ValueError("You can't provide both schema and table, because the table you provide"
                             "should already contain the schema.")
        if not schema and not table:
            raise ValueError("You must provide either schema or table.")

        if isinstance(schema, Path):
            schema = json.loads(schema.read_text())

        if table is None:
            table = Table(table_id, schema=schema)
            table.time_partitioning = TimePartitioning()

        self.logger.info(f'CREATING TABLE FROM SCHEMA: {table.schema}')

        self.bigquery_client.create_table(table)

    def insert(
            self,
            table_id: str,
            records: typing.Union[typing.List[dict], Path]):
        self.logger.info('INSERTING RECORDS TO TABLE: %s', table_id)
        table = self.bigquery_client.get_table(table_id)
        if isinstance(records, Path):
            with open(records, 'r') as f:
                records = json.loads(f.read())
        errors = self.bigquery_client.insert_rows(table, records)
        if errors:
            raise ValueError(errors)

    def _query(self, sql, job_config=None):
        self.logger.info('COLLECTING DATA: %s', sql)
        if job_config:
            return self.bigquery_client.query(sql, job_config=job_config)
        else:
            return self.bigquery_client.query(sql)

    @staticmethod
    def _convert_to_humanbytes(size_in_bytes):
        size = float(size_in_bytes)
        power = 2 ** 10
        tera = 2 ** 40
        tb_cost = 5
        n = 0
        power_labels = {0: ' B', 1: ' KB', 2: ' MB', 3: ' GB', 4: ' TB', 5: ' PB'}
        cost = str(round(size / tera * tb_cost, 2)) + ' USD'
        while size > power:
            size /= power
            n += 1
        return {'size': str(round(size, 2)) + power_labels[n],
                'cost': cost}


def create_dataset(dataset_name, bigquery_client, location=DEFAULT_LOCATION):
    from google.cloud import bigquery
    dataset = bigquery.Dataset('{project_id}.{dataset_name}'.format(
        project_id=bigquery_client.project,
        dataset_name=dataset_name))
    dataset.location = location
    return bigquery_client.create_dataset(dataset, exists_ok=True)


def random_uuid(suffix=''):
    return uuid.uuid1().hex + suffix


def create_bigquery_client(project_id, credentials, location):
    from google.cloud import bigquery
    return bigquery.Client(
        project=project_id,
        credentials=credentials,
        location=location)


def create_dataset_manager(
        project_id,
        runtime,
        dataset_name=None,
        internal_tables=None,
        external_tables=None,
        extras=None,
        credentials=None,
        location=DEFAULT_LOCATION,
        logger=None) -> typing.Tuple[str, PartitionedDatasetManager]:
    """
    Dataset manager factory.
    If dataset does not exist then it will also create dataset with given name.

    :param dataset_name: string dataset name(not dataset id). If not provided, dataset_name will be random string.
    :param internal_tables: list of dataset table names that are gonna be available during processing.
    :param external_tables: dict where key is table alias and value is full table ID.
    :param extras: dict with custom parameters that will be available inside templates.
    :param runtime: string determine partition that will be used for write operations.
     runtime is also available inside templates as {dt}.
     runtime can be either full datetime 'YYYY-MM-DD hh:mm:ss' or just date 'YYYY-MM-DD'.
    :param credentials: google.auth.credentials.Credentials instance. If empty, default credentials on the machine will be used.
    :param project_id: string full project id where dataset being processed is available.
    :param location: string location of dataset will be used to create datasets, tables, jobs, etc. EU by default.
    :param logger: custom logger.
    :return: tuple (full dataset ID, dataset manager).
    """
    dataset_name = dataset_name or random_uuid(suffix='_test_case')
    internal_tables = internal_tables or {}
    external_tables = external_tables or {}
    extras = extras or {}
    if logger is None:
        logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)
        logger = logging.getLogger(__name__)

    client = create_bigquery_client(project_id, credentials, location)
    dataset = create_dataset(dataset_name, client, location)

    core_dataset_manager = DatasetManager(client, dataset, logger)
    templated_dataset_manager = TemplatedDatasetManager(core_dataset_manager, internal_tables, external_tables, extras, runtime)
    return dataset.full_dataset_id.replace(':', '.'), PartitionedDatasetManager(templated_dataset_manager, get_partition_from_run_datetime_or_none(runtime))



"""Compiles, reads, validates `requirements.txt` files.
"""

import re
import typing
import logging
import tempfile
import textwrap
import hashlib
import subprocess

from pathlib import Path
from typing import Dict, List

import bigflow.commons as bf_commons


logger = logging.getLogger(__name__)


def pip_compile(
    requiremenets: Path,
    *,
    dry_run=False,
    verbose=False,
    upgrade=False,
    upgrade_package="",
    prereleases=False,
    rebuild=False,
    extra_args=(),
):
    """Wraps 'pip-tools' command. Include hash of source file into the generated one."""

    requirements_txt = requiremenets.with_suffix(".txt")
    requirements_in = requiremenets.with_suffix(".in")
    logger.info("Compile requirements file %s ...", requirements_in)

    with tempfile.NamedTemporaryFile('w+t', prefix=f"{requirements_in.stem}-", suffix=".txt", delete=False) as txt_file:
        txt_path = Path(txt_file.name)

        if requirements_txt.exists() and not rebuild:
            txt_path.write_bytes(requirements_txt.read_bytes())

        bf_commons.run_process([
            "pip-compile",
            "--no-header",
            *(["-o", txt_path] if not dry_run else []),
            *(["--dry-run"] if dry_run else []),
            *(["--rebuild"] if rebuild else []),
            *(["--upgrade"] if upgrade else []),
            *(["--pre"] if prereleases else []),
            *(["--upgrade-package", upgrade_package] if upgrade_package else []),
            *(["-v"] if verbose else ["-q"]),
            *extra_args,
            str(requirements_in),
        ], check=True)

        reqs_content = txt_path.read_text()

    if dry_run:
        return

    source_hash = compute_requirements_in_hash(requirements_in)
    with open(requirements_txt, 'w+t') as out:
        logger.info("Write pip requirements file: %s", requirements_txt)
        out.write(textwrap.dedent(f"""\
            # *** autogenerated: don't edit ***
            # $source-hash: {source_hash}
            # $source-file: {requirements_in}
            #
            # run 'bigflow build-requirements {requirements_in}' to update this file

        """))
        out.write(reqs_content)


def detect_piptools_source_files(requirements_dir: Path) -> typing.List[Path]:
    in_files = list(requirements_dir.glob("*.in"))

    manifest_file = requirements_dir / "MANIFEST.in"
    if manifest_file in in_files:
        in_files.remove(manifest_file)

    logger.debug("Found %d *.in files: %s", len(in_files), in_files)
    return in_files


def maybe_recompile_requirements_file(requirements_txt: Path) -> bool:
    # Some users keeps extra ".txt" files in the same directory.
    # Check if thoose files needs to be recompiled & then print a warning.
    for fin in detect_piptools_source_files(requirements_txt.parent):
        if fin.stem != requirements_txt.stem:
            check_requirements_needs_recompile(fin.with_suffix(".txt"))

    if check_requirements_needs_recompile(requirements_txt):
        pip_compile(requirements_txt)
        return True
    else:
        logger.debug("File %s is fresh", requirements_txt)
        return False


def _collect_all_input_files_content(requirements_in: Path):
    logger.debug("Scan all requiremenets.in-like files: %s", requirements_in)
    c = requirements_in.read_text()
    yield c
    for include in re.findall(r"\s*-r\s+(.*)", c):
        include: str
        fn = include.split("#", 1)[0].strip()
        yield from _collect_all_input_files_content(requirements_in.parent / fn)


def compute_requirements_in_hash(requirements_in: Path):
    logger.debug("Calculate hash of %s", requirements_in)
    algorithm = 'sha256'
    h = hashlib.new(algorithm)
    for c in _collect_all_input_files_content(requirements_in):
        h.update(c.encode())
    return algorithm + ":" + h.hexdigest()


def check_requirements_needs_recompile(requiremenets: Path) -> bool:
    """Checks if `requirements.{in,txt}` needs to be recompiled by `pip_compile()`"""

    requirements_txt = requiremenets.with_suffix(".txt")
    requirements_in = requiremenets.with_suffix(".in")
    logger.debug("Check if file %s should be recompiled", requirements_txt)

    if not requirements_in.exists():
        logger.debug("No file %s - pip-tools is not used", requirements_in)
        return False

    if not requirements_txt.exists():
        logger.debug("File %s does not exist - need to be compiled by 'pip-compile'", requirements_txt)
        return True

    requirements_txt_content = requirements_txt.read_text()
    hash1 = compute_requirements_in_hash(requirements_in)
    same_hash = hash1 in requirements_txt_content

    if same_hash:  # dirty but works ;)
        logger.debug("Don't need to compile %s file", requirements_txt)
        return False
    else:
        logger.warning("File %s needs to be recompiled with 'bigflow build-requirements' command", requirements_txt)
        return True


def read_requirements(requirements_path: Path, recompile_check=True) -> List[str]:
    """Reads and parses 'requirements.txt' file.

    Returns list of requirement specs, skipping comments and empty lines and pip directives.
    """

    if recompile_check and check_requirements_needs_recompile(requirements_path):
        raise ValueError("Requirements needs to be recompiled with 'pip-tools'")

    result: List[str] = []
    with open(requirements_path) as base_requirements:
        for line in base_requirements:
            line = line.split("#", 1)[0].strip()
            if line.startswith("-r "):
                subrequirements_file_name = line.replace("-r ", "")
                subrequirements_path = requirements_path.parent / subrequirements_file_name
                result.extend(read_requirements(subrequirements_path, recompile_check=False))
            elif line.startswith("--"):
                logger.debug("skip requirements line %r", line)
            elif line:
                result.append(line)

    return result


def generate_pinfile(
    requirements_path: Path,
    pins_file_in: Path,
    resolve_pins: typing.Callable[[], List[str]],
):
    requirements_txt = requirements_path.with_suffix(".txt")
    requirements_in = requirements_path.with_suffix(".in")

    logger.info("Clean pins file %s", pins_file_in)
    pins_file_in.write_text("# autocleaned ...")
    current_requirements_txt = requirements_txt.read_text() if requirements_txt.exists() else None
    pip_compile(requirements_in, rebuild=True)

    pins = resolve_pins()
    logger.info("Found %d pins: %s", len(pins), ", ".join(pins))

    # keep old .txt - preserve upgrades
    if current_requirements_txt:
        requirements_txt.write_text(current_requirements_txt)
    else:
        requirements_txt.unlink()

    _include_pinsfile_into_requirements(pins_file_in, requirements_in)
    bad_pins, all_pins = _try_incrementally_add_pins(pins_file_in, requirements_in, pins)

    logger.info("\nWrite %s", pins_file_in)
    pins_file_in.write_text("\n".join([
        f"# *** autogenerated ***",
        *all_pins,
    ]))

    logger.info("Recompile requirements...")
    pip_compile(requirements_in)

    if bad_pins:
        libs_list = "\n".join(f" - {x}" for x in sorted(bad_pins))
        logger.error("Failed to pin some libraries: \n%s", libs_list)
        logger.error("You may try to remove unused dependencies, upgrade beam or bigflow.")

    logger.info("Done")


def _try_incrementally_add_pins(pins_file_in, requirements_in, pins):
    bad_pins = []
    all_pins = []

    for i, pin in enumerate(pins):
        logger.info("\n(%d/%d) Trying to pin %r ...", i + 1, len(pins), pin)
        logger.info("Temporarily add %r to %s", pin, pins_file_in)
        pins_file_in.write_text("\n".join([
            "# *** autogenerated - partial! ***",
            *all_pins,
            pin,
            "# ...",
        ]))
        try:
            pip_compile(requirements_in, dry_run=True)
        except subprocess.CalledProcessError:
            logger.error("FAIL, revert")
            all_pins.append(f"## {pin}  # CONFLICT")
            bad_pins.append(pin)
        else:
            logger.info("OK, keep it")
            all_pins.append(pin)

    return bad_pins, all_pins


def _include_pinsfile_into_requirements(pins_file_in, requirements_in):
    req = requirements_in.read_text()
    m = re.search(rf"\s+-r\s+{re.escape(pins_file_in.name)}\s*#.*$", req)
    if m:
        logger.info("Pins file %s is already included into %s", pins_file_in, requirements_in)
    else:
        logger.info("Include pins file %s into %s", pins_file_in, requirements_in)
        relative_pins_path = pins_file_in.relative_to(requirements_in.parent)
        requirements_in.write_text(f"{req}\n-r {relative_pins_path} #  added by `bigflow`")
